import json
import re
from sqlalchemy import String, Enum, Date
from sqlalchemy.orm import MANYTOONE, ONETOMANY
from sqlalchemy.sql.elements import TextClause

from jet_bridge_base import status, settings
from jet_bridge_base.db import get_mapped_base, get_request_connection, request_connection_cache, MODEL_DESCRIPTIONS_RESPONSE_CACHE_KEY, MODEL_DESCRIPTIONS_HASH_CACHE_KEY
from jet_bridge_base.db_types import inspect_uniform
from jet_bridge_base.db_types.mongo import MongoColumn
from jet_bridge_base.models.model_relation_override import ModelRelationOverrideModel
from jet_bridge_base.responses.base import Response
from jet_bridge_base.store import store
from jet_bridge_base.models import data_types
from jet_bridge_base.permissions import HasProjectPermissions
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.serializers.model_description import ModelDescriptionSerializer
from jet_bridge_base.views.base.api import APIView
from jet_bridge_base.utils.common import merge, get_set_first
from jet_bridge_base.utils.crypt import get_sha256_hash
from jet_bridge_base.utils.db_types import sql_to_map_type, sql_to_db_type
from jet_bridge_base.utils.relations import relationship_direction_to_str
from jet_bridge_base.utils.tables import get_table_name


def uniform_to_map_type(column):
    try:
        if isinstance(column, MongoColumn):
            return column.type
        else:
            return sql_to_map_type(column.type)
    except:
        return 'NullType'


def uniform_to_db_type(column):
    try:
        if isinstance(column, MongoColumn):
            return column.type
        else:
            return sql_to_db_type(column.type)
    except:
        return 'NullType'


def is_column_has_default(column):
    return column.autoincrement or column.default or column.server_default


def is_column_optional(column):
    return is_column_has_default(column) or column.nullable


def is_column_required(column, primary_key_auto):
    if primary_key_auto:
        return True
    else:
        return not is_column_optional(column)


def map_column_default(column):
    if column.server_default is not None:
        if hasattr(column.server_default, 'arg') and isinstance(column.server_default.arg, TextClause):
            value = column.server_default.arg.text
            if value.lower() == 'now()':
                return {
                    'default_type': 'datetime_now'
                }
            elif value.lower() == 'uuid_generate_v4()':
                return {
                    'default_type': 'uuid'
                }
            elif value.lower() == 'true':
                return {
                    'default_type': 'value',
                    'default_value': True
                }
            elif value.lower() == 'false':
                return {
                    'default_type': 'value',
                    'default_value': False
                }

            value_regex = re.search("^'(?P<value>.+)'::(?P<type>.+)$", value)
            if value_regex:
                match = value_regex.groupdict()
                return {
                    'default_type': 'value',
                    'default_value': match['value']
                }

            nextval_regex = re.search("^nextval\((?P<value>.+)\)$", value)
            if nextval_regex:
                match = nextval_regex.groupdict()
                return {
                    'default_type': 'sequence',
                    'default_value': match['value']
                }


def map_column(metadata, column, editable, primary_key_auto):
    params = {}
    data_source_field = None
    data_source_name = None
    data_source_params = None
    data_source_order_after = None
    data_source_hidden = None

    map_type = uniform_to_map_type(column)
    db_type = uniform_to_db_type(column)

    if column.foreign_keys:
        foreign_key = next(iter(column.foreign_keys))
        map_type = data_types.FOREIGN_KEY
        params['related_model'] = {
            'model': get_table_name(metadata, foreign_key.column.table)
        }

        table_primary_keys = foreign_key.column.table.primary_key.columns.keys()
        table_primary_key = table_primary_keys[0] if len(table_primary_keys) > 0 else None

        if not table_primary_key or foreign_key.column.name != table_primary_key:
            params['custom_primary_key'] = foreign_key.column.name

    if isinstance(column.type, Date):
        params['date'] = True
        params['time'] = False

    if isinstance(column.type, Enum):
        params['options'] = map(lambda x: {'name': x, 'value': x}, column.type.enums)

    if isinstance(column.type, String):
        params['length'] = column.type.length

    try:
        from geoalchemy2 import types
        if isinstance(column.type, types.Geography):
            params['output_format'] = 'postgresql'
    except ImportError:
        pass

    required = is_column_required(column, primary_key_auto)

    if column.comment:
        try:
            data_source_meta = json.loads(column.comment)

            if not isinstance(data_source_meta, dict):
                raise ValueError

            meta_field = data_source_meta.get('field')
            meta_name = data_source_meta.get('name')
            meta_params = data_source_meta.get('params')
            meta_order_after = data_source_meta.get('order_after')
            meta_hidden = data_source_meta.get('hidden')

            if meta_field is not None:
                data_source_field = meta_field
            if meta_name is not None:
                data_source_name = meta_name
            if meta_params is not None and isinstance(meta_params, dict):
                data_source_params = meta_params
            if meta_order_after is not None:
                data_source_order_after = meta_order_after
            if meta_hidden is not None:
                data_source_hidden = meta_hidden
        except ValueError:
            pass

    result = {
        'name': column.name,
        'db_column': column.name,
        'field': map_type,
        'db_field': db_type,
        'filterable': True,
        'required': required,
        'null': column.nullable,
        'editable': editable,
        'params': params,
        'data_source_field': data_source_field,
        'data_source_name': data_source_name,
        'data_source_params': data_source_params,
        'data_source_order_after': data_source_order_after,
        'data_source_hidden': data_source_hidden
    }

    server_default = map_column_default(column)
    if server_default:
        result['default_type'] = server_default['default_type']
        if 'default_value' in server_default:
            result['default_value'] = server_default['default_value']

    return result


def is_column_primary_key_auto(primary_key, primary_key_auto, column):
    return primary_key is not None and column.name == primary_key.name and primary_key_auto


def map_table_column(metadata, table, column, editable):
    primary_key = table.primary_key.columns[0]
    primary_key_auto = getattr(table, '__jet_auto_pk__', False) if table is not None else None

    return map_column(
        metadata,
        column,
        editable,
        is_column_primary_key_auto(primary_key, primary_key_auto, column)
    )


def map_relationship(metadata, relationship):
    local_field = get_set_first(relationship.local_columns)
    related_field = get_set_first(relationship.remote_side)
    direction = relationship_direction_to_str(relationship.direction)

    result = {
        'name': relationship.key,
        'direction': direction,
        'local_field': local_field.name,
        'related_model': get_table_name(metadata, relationship.target),
        'related_field': related_field.name
    }

    return result


def map_relationship_override(override):
    result = {
        'name': override.name,
        'direction': override.direction,
        'local_field': override.local_field,
        'related_model': override.related_model,
        'related_field': override.related_field
    }

    return result


# def map_relation(relation):
#     field = None
#
#     if relation.direction == ONETOMANY:
#         field = 'ManyToOneRel'
#
#     return {
#         'name': relation.key,
#         'related_model': {
#             'model': relation.table.name
#         },
#         'field': field,
#         'related_model_field': relation.primaryjoin.right.name,
#         'through': None
#     }
#
# def table_relations(mapper):
#     return list(map(map_relation, filter(lambda x: x.direction == ONETOMANY and hasattr(x, 'table'), mapper.relationships)))
#
# def table_m2m_relations(mapper):
#     result = []
#     name = mapper.selectable.fullname
#
#     for relation in mapper.relationships:
#         if relation.direction != ONETOMANY or not hasattr(relation, 'table'):
#             continue
#
#         m2m_relationships = relation.mapper.relationships.values()
#
#         if len(m2m_relationships) != 2:
#             continue
#
#         if len(relation.table.columns) > 5:
#             continue
#
#         self_relationship = m2m_relationships[1] if m2m_relationships[1].table.fullname == name else \
#         m2m_relationships[0]
#         other_relationship = m2m_relationships[0] if self_relationship == m2m_relationships[1] else \
#         m2m_relationships[1]
#
#         result.append({
#             'name': 'M2M {} {}'.format(self_relationship.table.fullname, other_relationship.table.fullname),
#             'related_model': {
#                 'model': other_relationship.table.fullname
#             },
#             'field': 'ManyToManyField',
#             'related_model_field': self_relationship.table.fullname,
#             'through': {'model': relation.table.fullname}
#         })
#
#     return result


def map_table(MappedBase, cls, relationships_overrides, hidden):
    mapper = inspect_uniform(cls)
    table = mapper.tables[0]
    name = get_table_name(MappedBase.metadata, table)
    primary_key = mapper.primary_key[0]
    primary_key_auto = getattr(table, '__jet_auto_pk__', False) if table is not None else None
    is_view = getattr(table, '__jet_is_view__', False) if table is not None else False
    non_editable = []
    model_relationships_overrides = relationships_overrides.get(name)

    from jet_bridge_base.configuration import configuration
    additional = configuration.get_model_description(name)

    data_source_name = None
    data_source_name_plural = None
    data_source_order_after = None
    data_source_hidden = None

    if table.comment:
        try:
            data_source_meta = json.loads(table.comment)

            if not isinstance(data_source_meta, dict):
                raise ValueError

            meta_name = data_source_meta.get('name')
            meta_name_plural = data_source_meta.get('name_plural')
            meta_order_after = data_source_meta.get('order_after')
            meta_hidden = data_source_meta.get('hidden')

            if meta_name is not None:
                data_source_name = meta_name
            if meta_name_plural is not None:
                data_source_name_plural = meta_name_plural
            if meta_order_after is not None:
                data_source_order_after = meta_order_after
            if meta_hidden is not None:
                data_source_hidden = meta_hidden
        except ValueError:
            pass

    result = {
        'model': name,
        'db_table': name,
        'fields': list(map(lambda x: map_column(
            MappedBase.metadata,
            x,
            x.name not in non_editable,
            is_column_primary_key_auto(primary_key, primary_key_auto, x)
        ), mapper.columns)),
        'relations': sorted(
            list(map(lambda x: map_relationship(MappedBase.metadata, x), filter(lambda x: x.direction in [MANYTOONE, ONETOMANY], mapper.relationships))),
            key=lambda x: x['name']
        ),
        'relation_overrides': sorted(
            list(map(lambda x: map_relationship_override(x), model_relationships_overrides)),
            key=lambda x: x['name']
        ) if model_relationships_overrides else None,
        'hidden': name in hidden or name in configuration.get_hidden_model_description(),
        # 'relations': table_relations(mapper) + table_m2m_relations(mapper),
        'primary_key_field': primary_key.name if primary_key is not None else None,
        'primary_key_auto': primary_key_auto,
        'is_view': is_view,
        'data_source_name': data_source_name,
        'data_source_name_plural': data_source_name_plural,
        'data_source_order_after': data_source_order_after,
        'data_source_hidden': data_source_hidden
    }

    if additional:
        result = merge(result, additional)

    return result


class ModelDescriptionView(APIView):
    serializer_class = ModelDescriptionSerializer
    permission_classes = (HasProjectPermissions,)

    def get_queryset(self, request):
        hidden = ['__jet__token']
        MappedBase = get_mapped_base(request)
        relationships_overrides = {}
        draft = bool(request.get_argument('draft', False))

        if store.is_ok():
            connection = get_request_connection(request)

            with store.session() as session:
                overrides = session.query(ModelRelationOverrideModel).filter(
                    ModelRelationOverrideModel.connection_id == connection['id'],
                    draft == draft
                ).all()

                for override in overrides:
                    if override.model not in relationships_overrides:
                        relationships_overrides[override.model] = []
                    relationships_overrides[override.model].append(override)

        return sorted(
            list(map(lambda x: map_table(MappedBase, x, relationships_overrides, hidden), MappedBase.classes)),
            key=lambda x: x['model']
        )

    def get(self, request, *args, **kwargs):
        queryset = self.get_queryset(request)
        serializer = self.serializer_class(instance=queryset, many=True)
        cid = request.get_argument('cid', None)
        client_cache_enabled = cid is not None

        with request_connection_cache(request) as cache:
            rendered_data = cache.get(MODEL_DESCRIPTIONS_RESPONSE_CACHE_KEY) \
                if settings.CACHE_MODEL_DESCRIPTIONS else None

            if rendered_data is not None:
                if client_cache_enabled:
                    rendered_data_hash = cache.get(MODEL_DESCRIPTIONS_HASH_CACHE_KEY, get_sha256_hash(rendered_data))
                else:
                    rendered_data_hash = None

                if client_cache_enabled:
                    not_modified_response = self.get_not_modified_response(request, rendered_data_hash)
                    if not_modified_response:
                        return not_modified_response

                response = JSONResponse(rendered_data=rendered_data)

                if client_cache_enabled:
                    self.set_response_cache_headers(response, rendered_data_hash)

                return response
            else:
                response = JSONResponse(serializer.representation_data)
                rendered_data = response.render()

                cache[MODEL_DESCRIPTIONS_RESPONSE_CACHE_KEY] = rendered_data

                if client_cache_enabled:
                    rendered_data_hash = get_sha256_hash(rendered_data)
                    cache[MODEL_DESCRIPTIONS_HASH_CACHE_KEY] = rendered_data_hash

                    not_modified_response = self.get_not_modified_response(request, rendered_data_hash)
                    if not_modified_response:
                        return not_modified_response

                    self.set_response_cache_headers(response, rendered_data_hash)

                return response

    def set_response_cache_headers(self, response, rendered_data_hash):
        response.headers['Cache-Control'] = 'no-cache'
        response.headers['ETag'] = '"%s"' % rendered_data_hash

    def get_not_modified_response(self, request, rendered_data_hash):
        if_none_match = request.headers.get('IF_NONE_MATCH')

        if if_none_match is not None and '"%s"' % rendered_data_hash == if_none_match:
            return Response(status=status.HTTP_304_NOT_MODIFIED)



from jet_bridge_base.db_types import queryset_aggregate
from jet_bridge_base.filters.char_filter import CharFilter
from jet_bridge_base.filters.filter import EMPTY_VALUES


class ModelAggregateFilter(CharFilter):

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        return queryset_aggregate(self.model, qs, value)



from sqlalchemy.orm import MANYTOONE, ONETOMANY


def relationship_direction_to_str(direction):
    if direction == MANYTOONE:
        return 'MANYTOONE'
    elif direction == ONETOMANY:
        return 'ONETOMANY'
    else:
        return str(direction)


def parse_relationship_direction(direction):
    if direction == 'MANYTOONE':
        return MANYTOONE
    elif direction == 'ONETOMANY':
        return ONETOMANY



from .mongo_declarative_meta import MongoDeclarativeMeta


class MongoBaseClasses(dict):
    def __init__(self, tables, *args, **kwargs):
        super(MongoBaseClasses, self).__init__(*args, **kwargs)

        for table in tables:
            self[table.name] = MongoDeclarativeMeta(table)

    def __iter__(self):
        for value in self.values():
            yield value


class MongoBase(object):
    def __init__(self, metadata):
        self.metadata = metadata
        self.classes = MongoBaseClasses(metadata.tables)



from jet_bridge_base import fields as fields_
from jet_bridge_base.serializers.serializer import Serializer


class ModelDescriptionFieldSerializer(Serializer):
    name = fields_.CharField()
    db_column = fields_.CharField()
    field = fields_.CharField()
    db_field = fields_.CharField()
    filterable = fields_.BooleanField()
    null = fields_.BooleanField()
    editable = fields_.BooleanField()
    params = fields_.JSONField()
    data_source_field = fields_.CharField(required=False)
    data_source_name = fields_.CharField(required=False)
    data_source_params = fields_.JSONField(required=False)
    data_source_order_after = fields_.CharField(required=False)
    data_source_hidden = fields_.BooleanField(required=False)
    verbose_name = fields_.CharField(required=False)
    required = fields_.BooleanField(required=False)
    default_type = fields_.CharField(required=False)
    default_value = fields_.RawField(required=False)
    length = fields_.IntegerField(required=False)
    related_model = fields_.JSONField(required=False)
    custom_primary_key = fields_.CharField(required=False)


class ModelDescriptionRelationSerializer(Serializer):
    name = fields_.CharField()
    direction = fields_.CharField()
    local_field = fields_.CharField()
    related_model = fields_.CharField()
    related_field = fields_.CharField()


class ModelDescriptionSerializer(Serializer):
    model = fields_.CharField()
    db_table = fields_.CharField()
    hidden = fields_.BooleanField()
    primary_key_field = fields_.CharField()
    primary_key_auto = fields_.BooleanField()
    is_view = fields_.BooleanField()
    fields = ModelDescriptionFieldSerializer(many=True)
    relations = ModelDescriptionRelationSerializer(many=True)
    relation_overrides = ModelDescriptionRelationSerializer(many=True, required=False)
    verbose_name = fields_.CharField(required=False)
    verbose_name_plural = fields_.CharField(required=False)
    display_field = fields_.CharField(required=False)
    default_order_by = fields_.CharField(required=False)
    data_source_name = fields_.CharField(required=False)
    data_source_name_plural = fields_.CharField(required=False)
    data_source_order_after = fields_.CharField(required=False)
    data_source_hidden = fields_.BooleanField(required=False)



import graphene

from jet_bridge_base.utils.gql import RawScalar

EXACT = 'exact'
GT = 'gt'
GTE = 'gte'
LT = 'lt'
LTE = 'lte'
ICONTAINS = 'icontains'
IN = 'in'
STARTS_WITH = 'istartswith'
ENDS_WITH = 'iendswith'
IS_NULL = 'isnull'
IS_EMPTY = 'isempty'
JSON_ICONTAINS = 'json_icontains'
COVEREDBY = 'coveredby'
DEFAULT_LOOKUP = EXACT

by_gql = {
    'eq': EXACT,
    'gt': GT,
    'gte': GTE,
    'lt': LT,
    'lte': LTE,
    'containsI': ICONTAINS,
    'in': IN,
    'startsWithI': STARTS_WITH,
    'endsWithI': ENDS_WITH,
    'isNull': IS_NULL,
    'isEmpty': IS_EMPTY,
    'jsonContainsI': JSON_ICONTAINS,
    'coveredBy': COVEREDBY
}

gql_scalar = {
    IN: graphene.List(RawScalar)
}

gql = dict(map(lambda x: (x[1], x[0]), by_gql.items()))



import datetime
import dateparser
import six

from jet_bridge_base.db import get_default_timezone
from jet_bridge_base.fields.field import Field


def get_timezone_from_str(value):
    if not value:
        return

    return datetime.datetime.strptime(value.replace(':', ''), '%z').tzinfo


def datetime_apply_default_timezone(value, request):
    if value.tzinfo is not None:
        return value

    session_timezone = request.session.info.get('_queries_timezone')
    if session_timezone:
        try:
            request_timezone = get_timezone_from_str(session_timezone)
            return value.replace(tzinfo=request_timezone)
        except:
            pass

    default_timezone = get_default_timezone(request) if request else None
    if default_timezone:
        return value.replace(tzinfo=default_timezone)

    return value


class DateTimeField(Field):
    field_error_messages = {
        'invalid': 'date has wrong format'
    }

    def to_internal_value_item(self, value):
        if value is None:
            return
        value = six.text_type(value).strip()

        try:
            result = dateparser.parse(value)
        except ValueError:
            result = None

        if result is None:
            self.error('invalid')

        return result

    def to_representation_item(self, value):
        if value is None:
            return

        request = self.context.get('request')
        value = datetime_apply_default_timezone(value, request)

        return value.isoformat()



from __future__ import absolute_import
import json
from six import string_types

from jet_bridge_base.fields.field import Field


class ArrayField(Field):
    field_error_messages = {
        'invalid': 'not a valid array'
    }

    def to_internal_value_item(self, value):
        internal_value = None

        if isinstance(value, string_types):
            try:
                internal_value = json.loads(value)
            except ValueError:
                self.error('invalid')
        else:
            internal_value = value

        if not isinstance(internal_value, list):
            raise ValueError

        return internal_value

    def to_representation_item(self, value):
        return value



from datetime import tzinfo, timedelta


class FixedOffsetTimezone(tzinfo):
    def __init__(self, offset):
        self.offset = offset
        self.name = 'Etc/GMT%+d' % (offset.total_seconds() / 60 / 60)

    def tzname(self, dt):
        return self.name

    def utcoffset(self, dt):
        return self.offset

    def dst(self, dt):
        return timedelta(0)

    def __repr__(self):
        return 'FixedOffsetTimezone(name={}, offset={})'.format(self.name, self.offset)




def action(methods=None, detail=False):
    if methods is None:
        methods = ['get']
    else:
        methods = [method.lower() for method in methods]

    def decorator(func):
        func.bind_to_methods = methods
        func.detail = detail
        return func
    return decorator



import graphene


class RawScalar(graphene.Scalar):
    @staticmethod
    def serialize(value):
        return value

    @staticmethod
    def parse_literal(node, _variables=None):
        return node.value

    @staticmethod
    def parse_value(value):
        return value



from __future__ import absolute_import
import json

from jet_bridge_base import encoders
from jet_bridge_base.responses.base import Response


class JSONResponse(Response):
    encoder_class = encoders.JSONEncoder

    def __init__(self, *args, **kwargs):
        self.rendered_data = kwargs.pop('rendered_data', None)
        super(JSONResponse, self).__init__(*args, **kwargs)

    def default_headers(self):
        return {'Content-Type': 'application/json'}

    def render(self):
        if self.rendered_data is not None:
            return self.rendered_data

        if self.data is None:
            return

        self.rendered_data = json.dumps(
            self.data,
            cls=self.encoder_class
        )
        return self.rendered_data



from fnmatch import fnmatch
import requests

from jet_bridge_base import settings
from jet_bridge_base.configuration import configuration
from jet_bridge_base.sentry import sentry_controller
from jet_bridge_base.utils.conf import get_conf


def track_database(conf):
    if not settings.TRACK_DATABASES_ENDPOINT:
        return

    track_databases = list(filter(lambda x: x != '', map(lambda x: x.lower().strip(), settings.TRACK_DATABASES.split(','))))
    hostname = '{}:{}'.format(conf.get('host', ''), conf.get('port', '')).lower()

    if not any(map(lambda x: fnmatch(hostname, x), track_databases)):
        return

    headers = {}
    data = {
        'databaseName': conf.get('name', ''),
        'databaseSchema': conf.get('schema', ''),
        'databaseHost': conf.get('host', ''),
        'databasePort': conf.get('port', '')
    }

    if settings.TRACK_DATABASES_AUTH:
        headers['Authorization'] = settings.TRACK_DATABASES_AUTH

    try:
        r = requests.post(settings.TRACK_DATABASES_ENDPOINT, headers=headers, json=data)
        success = 200 <= r.status_code < 300

        if not success:
            error = 'TRACK_DATABASE request error: {} {} {}'.format(r.status_code, r.reason, r.text)
            sentry_controller.capture_message(error)
    except Exception as e:
        sentry_controller.capture_exception(e)


def track_database_async(request):
    if not settings.TRACK_DATABASES_ENDPOINT:
        return

    conf = get_conf(request)
    configuration.run_async(track_database, conf)



import json
import re
import pymongo
from bson import ObjectId

from .common import mongo_inspect
from .mongo_operator import MongoOperator
from .mongo_record import MongoRecord
from .mongo_desc import MongoDesc


class MongoQueryset(object):
    select = None
    whereclause = None
    _joins = None
    _search = None
    _offset = None
    _limit = None
    _sort = None

    def __init__(self, session, name, select=None, whereclause=None, joins=None, search=None, offset=None, limit=None, sort=None):
        self.session = session
        self.db = session.db
        self.name = name
        self.query = self.db[self.name]
        self.select = select
        self.whereclause = whereclause
        self._joins = joins
        self._search = search
        self._offset = offset
        self._limit = limit
        self._sort = sort

    def to_internal_value(self, value, column):
        if isinstance(value, bytes) and column.params.get('type') == 'object_id':
            return ObjectId(value)
        else:
            return value

    def get_column_path(self, column):
        return '{}.{}'.format(column.table.name, column.name) if column.table.name != self.name else column.name

    def get_empty(self):
        return {
            '_id': {'$exists': False}
        }

    def map_operator(self, arg):
        acc = {}

        if isinstance(arg, MongoOperator):
            if arg.operator == '__eq__':
                column_path = self.get_column_path(arg.lhs)
                value = self.to_internal_value(arg.rhs, arg.lhs)
                acc[column_path] = value
            elif arg.operator == '__gt__':
                column_path = self.get_column_path(arg.lhs)
                value = self.to_internal_value(arg.rhs, arg.lhs)
                acc[column_path] = {'$gt': value}
            elif arg.operator == '__ge__':
                column_path = self.get_column_path(arg.lhs)
                value = self.to_internal_value(arg.rhs, arg.lhs)
                acc[column_path] = {'$gte': value}
            elif arg.operator == '__lt__':
                column_path = self.get_column_path(arg.lhs)
                value = self.to_internal_value(arg.rhs, arg.lhs)
                acc[column_path] = {'$lt': value}
            elif arg.operator == '__le__':
                column_path = self.get_column_path(arg.lhs)
                value = self.to_internal_value(arg.rhs, arg.lhs)
                acc[column_path] = {'$lte': value}
            elif arg.operator == 'exists':
                column_path = self.get_column_path(arg.lhs)
                value = self.to_internal_value(arg.rhs, arg.lhs)
                acc[column_path] = {'$exists': value}
            elif arg.operator == 'in':
                column_path = self.get_column_path(arg.lhs)
                value = self.to_internal_value(arg.rhs, arg.lhs)
                acc[column_path] = {'$in': value}
            elif arg.operator == 'or':
                acc['$or'] = list(map(lambda x: self.map_operator(x), arg.lhs))
            elif arg.operator == 'not':
                positive = self.map_operator(arg.lhs)
                for key, value in positive.items():
                    if isinstance(value, dict):
                        acc[key] = {'$not': value}
                    else:
                        acc[key] = {'$ne': value}
            elif arg.operator == 'json_icontains':
                acc.update(self.get_empty())
            elif arg.operator == 'ilike':
                column_path = self.get_column_path(arg.lhs)
                value = self.to_internal_value(arg.rhs, arg.lhs)
                acc[column_path] = {'$regex': self.get_regex_from_ilike(value)}

        return acc

    def get_regex_from_ilike(self, value):
        if not isinstance(value, str):
            return None

        if value.startswith('%'):
            value = value[1:]
            prefix = ''
        else:
            prefix = '^'

        if value.endswith('%'):
            value = value[:-1]
            postfix = ''
        else:
            postfix = '$'

        return re.compile(prefix + re.escape(value) + postfix, re.IGNORECASE)

    def filter(self, *args):
        if len(args) == 0:
            return self

        result = self.clone()

        for arg in args:
            if not result.whereclause:
                result.whereclause = []
            result.whereclause.append(self.map_operator(arg))

        return result

    def join(self, model, operator):
        result = self.clone()

        result._joins = result._joins or []
        result._joins.append((model, operator))

        return result

    def search(self, search):
        result = self.clone()
        result._search = search
        return result

    def offset(self, offset):
        result = self.clone()
        result._offset = offset
        return result

    def limit(self, limit):
        result = self.clone()
        result._limit = limit
        return result

    def get_limit(self):
        return self._limit

    def order_by(self, *columns):
        if len(columns) == 0:
            return self

        result = self.clone()
        result._sort = result._sort or []

        for item in columns:
            if isinstance(item, MongoDesc):
                result._sort.append((item.column.name, pymongo.DESCENDING))
            else:
                result._sort.append((item.name, pymongo.ASCENDING))

        return result

    def get_order_by(self):
        return self._sort

    def clone(self):
        return MongoQueryset(self.session, self.name, select=self.select, whereclause=self.whereclause,
                             joins=self._joins, search=self._search, offset=self._offset, limit=self._limit,
                             sort=self._sort)

    def get_filters(self):
        filters = []

        if self.whereclause:
            filters.extend(self.whereclause)

        if self._search is not None:
            filters.append({'$text': {'$search': self._search}})

        if len(filters) > 1:
            return {'$and': filters}
        elif len(filters) == 1:
            return filters[0]
        else:
            return {}

    def get_aggregate_pipeline(self, limit=None):
        aggregate = []

        if self._joins:
            for model, operator in self._joins:
                mapper = mongo_inspect(model)
                aggregate.append({'$lookup': {
                    'from': mapper.selectable.name,
                    'localField': operator.rhs.name,
                    'foreignField': operator.lhs.name,
                    'as': mapper.selectable.name
                }})

        filters = self.get_filters()
        if filters:
            aggregate.append({'$match': filters})

        if self._sort:
            aggregate.append({'$sort': dict(self._sort)})

        if self._offset:
            aggregate.append({'$skip': self._offset})

        if limit is not None:
            aggregate.append({'$limit': limit})
        elif self._limit:
            aggregate.append({'$limit': self._limit})

        if self.select:
            aggregate.append({'$project': dict(map(lambda x: (x.name, 1), self.select))})

        return aggregate

    def all(self):
        pipeline = self.get_aggregate_pipeline()
        for data in self.query.aggregate(pipeline, allowDiskUse=True):
            record = MongoRecord(self.name, **data)
            self.session.bind_record(record)

            yield record

    def count(self):
        pipeline = self.get_aggregate_pipeline()
        pipeline.append({'$count': 'count'})

        try:
            result = self.query.aggregate(pipeline, allowDiskUse=True).next()
            return result['count']
        except StopIteration:
            return 0

    def estimated_document_count(self):
        return self.query.estimated_document_count()

    def first(self):
        pipeline = self.get_aggregate_pipeline(limit=1)

        try:
            data = self.query.aggregate(pipeline, allowDiskUse=True).next()
        except StopIteration:
            return None

        record = MongoRecord(self.name, **data)
        self.session.bind_record(record)

        return record

    def one(self):
        return self.first()

    def aggregate(self, pipeline):
        filters = self.get_filters()

        try:
            return self.query.aggregate([{'$match': filters}, pipeline]).next()
        except StopIteration:
            return {'aggregation': 0}

    def group(self, pipeline, sort):
        filters = self.get_filters()
        return list(self.query.aggregate([{'$match': filters}, pipeline, sort]))

    def __iter__(self):
        return self.all()

    def __repr__(self):
        return json.dumps({
            'whereclause': list(map(lambda x: repr(x), self.whereclause)) if self.whereclause else None,
            '_search': self._search,
            '_offset': self._offset,
            '_limit': self._limit,
            '_sort': self._sort,
        })



import os

from jet_bridge_base import settings
from jet_bridge_base.logger import logger
from jet_bridge_base.utils.conf import get_connection_id, get_connection_schema, get_connection_name, \
    get_metadata_file_path

from .mongo import mongo_dump_metadata_file, MongoMetadata
from .sql import sql_dump_metadata_file


def dump_metadata_file(conf, metadata):
    if isinstance(metadata, MongoMetadata):
        mongo_dump_metadata_file(conf, metadata)
    else:
        sql_dump_metadata_file(conf, metadata)


def remove_metadata_file(conf):
    if not settings.CACHE_METADATA:
        return

    connection_id = get_connection_id(conf)
    schema = get_connection_schema(conf)
    connection_name = get_connection_name(conf, schema)
    id_short = connection_id[:4]

    file_path = get_metadata_file_path(conf)

    if not os.path.exists(file_path):
        logger.info('[{}] Schema cache clear skipped (file not found) for "{}"'.format(id_short, connection_name))
        return

    try:
        os.unlink(file_path)
        logger.info('[{}] Schema cache cleared for "{}"'.format(id_short, connection_name))

        return file_path
    except Exception as e:
        logger.error('[{}] Schema cache clear failed for "{}"'.format(id_short, connection_name), exc_info=e)



from pymongo import MongoClient
from sqlalchemy import MetaData, inspection
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.sql.base import _bind_or_error

from jet_bridge_base.db_types.sql import sql_create_connection_engine, sql_get_tables
from jet_bridge_base.utils.conf import get_connection_schema


def discover_connection(conf, tunnel):
    bind = None
    client = None

    try:
        if conf.get('engine') == 'mongo':
            database_url = conf.get('url')
            database_name = conf.get('name')

            client = MongoClient(database_url)

            if database_name not in client.list_database_names():
                raise Exception('No such database found or database is empty: {}'.format(database_name))

            db = client[conf.get('name')]
            db.list_collection_names()

            return True
        else:
            bind = sql_create_connection_engine(conf, tunnel)

            Session = scoped_session(sessionmaker(bind=bind))
            session = Session()

            with session.connection():
                return True
    except Exception as e:
        raise e
    finally:
        if bind:
            bind.dispose()

        if client:
            client.close()

        if tunnel:
            tunnel.close()


def discover_tables(conf, tunnel):
    bind = None
    client = None

    try:
        if conf.get('engine') == 'mongo':
            database_url = conf.get('url')
            database_name = conf.get('name')

            client = MongoClient(database_url)

            if database_name not in client.list_database_names():
                raise Exception('No such database found or database is empty: {}'.format(database_name))

            db = client[conf.get('name')]

            return db.list_collection_names()
        else:
            schema = get_connection_schema(conf)
            bind = sql_create_connection_engine(conf, tunnel)

            Session = scoped_session(sessionmaker(bind=bind))
            session = Session()

            with session.connection() as connection:
                metadata = MetaData(schema=schema, bind=connection)

                if bind is None:
                    bind = _bind_or_error(metadata)

                with inspection.inspect(bind)._inspection_context() as insp:
                    if schema is None:
                        schema = metadata.schema

                    load, view_names = sql_get_tables(insp, metadata, bind, schema, foreign=True, views=True)
                    return load
    except Exception as e:
        raise e
    finally:
        if bind:
            bind.dispose()

        if client:
            client.close()

        if tunnel:
            tunnel.close()



from jet_bridge_base import status
from jet_bridge_base.exceptions.api import APIException


class NotFound(APIException):
    default_detail = 'Not Found'
    default_code = 'not_found'
    default_status_code = status.HTTP_404_NOT_FOUND



from jet_bridge_base.utils.common import CollectionDict

from .mongo_column import MongoColumn


class MongoTable(object):
    def __init__(self, name, columns=None, comment=None, schema=None):
        self.name = name
        self.columns = CollectionDict(map(lambda x: (
            x['name'],
            MongoColumn.deserialize(self, x)
        ), columns)) if columns else CollectionDict()
        self.comment = comment
        self.schema = schema

    def append_column(self, column):
        self.columns[column.name] = column

    @staticmethod
    def deserialize(obj):
        name = obj.pop('name')
        return MongoTable(name, **obj)

    def serialize(self):
        return {
            'name': self.name,
            'columns': list(map(lambda x: x.serialize(), self.columns.values())),
            'comment': self.comment,
            'schema': self.schema
        }

    def __repr__(self):
        return self.name



from jet_bridge_base.fields.field import Field


class RawField(Field):

    def to_internal_value_item(self, value):
        return value

    def to_representation_item(self, value):
        return value



import json
from datetime import datetime

from jet_bridge_base import settings
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.status import HTTP_400_BAD_REQUEST
from jet_bridge_base.utils.crypt import decrypt, get_sha256_hash
from jet_bridge_base.utils.token import parse_token, JWT_TOKEN_PREFIX, decode_jwt_token, decompress_permissions
from jet_bridge_base.views.base.api import BaseAPIView


class TokenInspectView(BaseAPIView):

    def get(self, request, *args, **kwargs):
        token_str = request.get_argument('authorization', default=None) or request.headers.get('AUTHORIZATION')
        if not token_str:
            return JSONResponse({'error': 'Token not specified'}, status=HTTP_400_BAD_REQUEST)

        token = parse_token(token_str)

        if not token:
            return JSONResponse({'error': 'Token parse failed'}, status=HTTP_400_BAD_REQUEST)

        response = {
            'token_type': token['type'],
            'token_value': token['value'],
            'token_params': token['params']
        }

        if token['type'] == JWT_TOKEN_PREFIX:
            jwt_value = decode_jwt_token(token['value'], verify_exp=False)

            response['jwt_data'] = {**jwt_value}

            if jwt_value:
                expire = datetime.utcfromtimestamp(jwt_value['exp'])

                response['jwt_data']['exp'] = expire.isoformat()
                response['jwt_data']['expired'] = datetime.utcnow() >= expire
                response['jwt_data']['exp_raw'] = jwt_value['exp']

                projects_extended = {}

                for project, user_permissions in jwt_value.get('projects', {}).items():
                    if 'permissions' in user_permissions:
                        permissions = decompress_permissions(user_permissions['permissions'])
                    else:
                        permissions = []

                    projects_extended[project] = {
                        **user_permissions,
                        'permissions': permissions,
                        'permissions_raw': user_permissions.get('permissions')
                    }

                response['jwt_data']['projects'] = projects_extended

        bridge_settings_str = request.get_argument('bridge_settings', default=None) or request.headers.get('X_BRIDGE_SETTINGS')
        if bridge_settings_str:
            response['bridge_settings_str'] = bridge_settings_str

            try:
                secret_key = settings.TOKEN.replace('-', '').lower()
                decrypted = decrypt(bridge_settings_str, secret_key)

                bridge_settings = json.loads(decrypted)
                bridge_settings_token = bridge_settings.get('token')

                response['bridge_settings'] = {
                    'token': bridge_settings_token,
                    'token_hash': get_sha256_hash(bridge_settings_token.replace('-', '').lower()),
                    'project': bridge_settings.get('project')
                }
            except Exception as e:
                response['bridge_settings_error'] = repr(e)

        return JSONResponse(response)



from jet_bridge_base.db import dispose_request_connection, get_request_connection
from jet_bridge_base.db_types import remove_metadata_file
from jet_bridge_base.permissions import HasProjectPermissions
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.utils.conf import get_conf
from jet_bridge_base.views.base.api import APIView


class ReloadView(APIView):
    permission_classes = (HasProjectPermissions,)

    def required_project_permission(self, request):
        return {
            'permission_type': 'project',
            'permission_object': 'project_settings',
            'permission_actions': ''
        }

    def post(self, request, *args, **kwargs):
        conf = get_conf(request)
        remove_metadata_file(conf)

        result = dispose_request_connection(request)
        get_request_connection(request)

        return JSONResponse({
            'success': result
        })



import json

from jet_bridge_base import status
from jet_bridge_base.db import get_mapped_base, get_engine, reload_request_mapped_base
from jet_bridge_base.encoders import JSONEncoder
from jet_bridge_base.exceptions.not_found import NotFound
from jet_bridge_base.exceptions.validation_error import ValidationError
from jet_bridge_base.permissions import HasProjectPermissions
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.serializers.table import TableSerializer
from jet_bridge_base.views.base.api import APIView
from jet_bridge_base.views.table_column import map_dto_column
from sqlalchemy import Table
from sqlalchemy.exc import InternalError


class TableView(APIView):
    permission_classes = (HasProjectPermissions,)

    def get_db(self, request):
        MappedBase = get_mapped_base(request)
        engine = get_engine(request)
        return MappedBase.metadata, engine

    def update_base(self, request):
        reload_request_mapped_base(request)

    def get_object(self, request):
        metadata, engine = self.get_db(request)
        pk = request.path_kwargs['pk']

        try:
            obj = list(filter(lambda x: x.name == pk, metadata.tables.values()))[0]
        except IndexError:
            raise NotFound

        self.check_object_permissions(request, obj)

        return obj

    def create(self, request, *args, **kwargs):
        serializer = TableSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        try:
            self.perform_create(request, serializer)
        except Exception as e:
            raise ValidationError(str(e))

        return JSONResponse(serializer.representation_data, status=status.HTTP_201_CREATED)

    def perform_create(self, request, serializer):
        data = serializer.validated_data
        metadata, engine = self.get_db(request)

        data_source_meta = {}
        data_source_name = data.get('data_source_name')
        data_source_name_plural = data.get('data_source_name_plural')

        if data_source_name:
            data_source_meta['name'] = data_source_name

        if data_source_name_plural:
            data_source_meta['name_plural'] = data_source_name_plural

        if len(data_source_meta.keys()):
            comment = json.dumps(data_source_meta, cls=JSONEncoder)
        else:
            comment = None

        columns = list(map(lambda x: map_dto_column(data['name'], x, metadata), data['columns']))
        table = Table(
            data['name'],
            metadata,
            *columns,
            comment=comment
        )

        try:
            table.create(bind=engine)

            metadata._set_parent(table)
            self.update_base(request)
        except Exception as e:
            metadata.remove(table)
            raise e

    def destroy(self, request, *args, **kwargs):
        instance = self.get_object(request)
        self.perform_destroy(request, instance)
        return JSONResponse(status=status.HTTP_204_NO_CONTENT)

    def perform_destroy(self, request, table):
        metadata, engine = self.get_db(request)

        try:
            table.drop(bind=engine)
        except InternalError as e:
            raise ValidationError(str(e))

        metadata.remove(table)
        self.update_base(request)



from jet_bridge_base import settings
from jet_bridge_base.db import get_connection_tunnel
from jet_bridge_base.db_types import discover_tables
from jet_bridge_base.exceptions.validation_error import ValidationError
from jet_bridge_base.permissions import HasProjectPermissions
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.utils.conf import get_conf
from jet_bridge_base.views.base.api import BaseAPIView


class DiscoverTableView(BaseAPIView):
    permission_classes = (HasProjectPermissions,)

    def required_project_permission(self, request):
        return {
            'permission_type': 'project',
            'permission_object': 'project_customization',
            'permission_actions': ''
        }

    def get(self, request, *args, **kwargs):
        conf = get_conf(request)

        try:
            tunnel = get_connection_tunnel(conf)
            tables = discover_tables(conf, tunnel)

            return JSONResponse({
                'tables': tables,
                'max_tables': settings.DATABASE_MAX_TABLES
            })
        except Exception as e:
            raise ValidationError(str(e))



from jet_bridge_base import fields
from jet_bridge_base.messages import get_handler
from jet_bridge_base.serializers.serializer import Serializer


class MessageSerializer(Serializer):
    name = fields.CharField()
    params = fields.JSONField(required=False)

    def save(self):
        handler = get_handler(self.validated_data['name'])
        if not handler:
            return
        return handler(self.validated_data['name'], self.validated_data.get('params', {}))



from jet_bridge_base import status
from jet_bridge_base.exceptions.validation_error import ValidationError


class RequestError(ValidationError):
    default_detail = 'invalid request'
    default_code = 'invalid_request'
    default_status_code = status.HTTP_400_BAD_REQUEST

    def __init__(self, request, *args, **kwargs):
        self.request = request
        super(RequestError, self).__init__(*args, **kwargs)



from jet_bridge_base.db_types import inspect_uniform
from jet_bridge_base.filters import lookups
from jet_bridge_base.filters.filter import Filter
from jet_bridge_base.filters.filter_for_dbfield import filter_for_column


class FilterClass(object):
    filters = []

    def __init__(self, *args, **kwargs):
        self.meta = getattr(self, 'Meta', None)
        if 'context' in kwargs:
            self.handler = kwargs['context'].get('handler', None)
        self.update_filters()

    def update_filters(self):
        filters = []

        if self.meta:
            if hasattr(self.meta, 'model'):
                Model = self.meta.model
                mapper = inspect_uniform(Model)
                columns = mapper.columns

                if hasattr(self.meta, 'fields'):
                    columns = filter(lambda x: x.name in self.meta.fields, columns)

                for column in columns:
                    item = filter_for_column(column)
                    for lookup in item['lookups']:
                        for exclude in [False, True]:
                            instance = item['filter_class'](
                                name=column.key,
                                column=column,
                                lookup=lookup,
                                exclude=exclude
                            )
                            filters.append(instance)

        declared_filters = filter(lambda x: isinstance(x[1], Filter), map(lambda x: (x, getattr(self, x)), dir(self)))

        for filter_name, filter_item in declared_filters:
            filter_item.name = filter_name
            filter_item.model = Model
            filter_item.handler = self.handler
            filters.append(filter_item)

        self.filters = filters

    def filter_queryset(self, request, queryset):
        def get_filter_value(name, filters_instance=None):
            return request.get_argument_safe(name, None)

        for item in self.filters:
            if item.name:
                argument_name = '{}__{}'.format(item.name, item.lookup)
                if item.exclude:
                    argument_name = 'exclude__{}'.format(argument_name)
                value = get_filter_value(argument_name, item)

                if value is None and item.lookup == lookups.DEFAULT_LOOKUP:
                    argument_name = item.name
                    if item.exclude:
                        argument_name = 'exclude__{}'.format(argument_name)
                    value = get_filter_value(argument_name, item)
            else:
                value = None

            queryset = item.filter(queryset, value)
        return queryset



from jet_bridge_base.exceptions.api import APIException
from jet_bridge_base.logger import logger


class SentryController(object):
    sentry_sdk = None

    def enable(self, dsn, release, tornado=False):
        try:
            import sentry_sdk

            integrations = []

            if tornado:
                try:
                    from sentry_sdk.integrations.tornado import TornadoIntegration
                    integrations.append(TornadoIntegration())
                except ImportError:
                    pass

            sentry_sdk.init(
                dsn=dsn,
                integrations=integrations,
                release=release,
                before_send=lambda event, hint: self.before_send(event, hint)
            )

            self.sentry_sdk = sentry_sdk
        except ImportError:
            self.sentry_sdk = None

    def before_send(self, event, hint):
        if 'exc_info' in hint:
            exc_type, exc_value, tb = hint['exc_info']
            if isinstance(exc_value, (APIException, KeyboardInterrupt)):
                return None
        if event.get('logger') == 'jet_bridge':
            return None
        return event

    def set_user(self, user):
        if not self.sentry_sdk:
            return

        self.sentry_sdk.set_user(user)

    def set_context(self, name, value):
        if not self.sentry_sdk:
            return

        self.sentry_sdk.set_context(name, value)

    def capture_exception(self, exc):
        logger.exception(exc)

        if not self.sentry_sdk:
            return

        self.sentry_sdk.capture_exception(exc)

    def capture_message(self, message):
        logger.error(message)

        if not self.sentry_sdk:
            return

        self.sentry_sdk.capture_message(message)


sentry_controller = SentryController()



from jet_bridge_base import status, fields
from jet_bridge_base.exceptions.validation_error import ValidationError
from jet_bridge_base.utils.exceptions import serialize_validation_error
from jet_bridge_base.views.model_description import inspect_uniform
from sqlalchemy import inspect
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.engine import Row

from jet_bridge_base.db import get_mapped_base
from jet_bridge_base.db_types import apply_default_ordering
from jet_bridge_base.exceptions.not_found import NotFound
from jet_bridge_base.filters.model import get_model_filter_class
from jet_bridge_base.filters.model_aggregate import ModelAggregateFilter
from jet_bridge_base.filters.model_group import ModelGroupFilter
from jet_bridge_base.permissions import HasProjectPermissions, ReadOnly
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.router import action
from jet_bridge_base.serializers.model import get_model_serializer
from jet_bridge_base.serializers.model_group import ModelGroupSerializer
from jet_bridge_base.serializers.reorder import get_reorder_serializer
from jet_bridge_base.serializers.reset_order import get_reset_order_serializer
from jet_bridge_base.utils.siblings import get_model_siblings
from jet_bridge_base.views.mixins.model import ModelAPIViewMixin


class ModelViewSet(ModelAPIViewMixin):
    permission_classes = (HasProjectPermissions, ReadOnly)
    track_queries = True

    def before_dispatch(self, request):
        super(ModelViewSet, self).before_dispatch(request)

    def on_finish(self):
        super(ModelViewSet, self).on_finish()

    def required_project_permission(self, request):
        model_name = self.get_model_name(request)
        return {
            'permission_type': 'model',
            'permission_object': model_name,
            'permission_actions': {
                'create': 'w',
                'update': 'w',
                'partial_update': 'w',
                'destroy': 'd',
                'retrieve': 'r',
                'list': 'r',
                'aggregate': 'r',
                'group': 'r',
                'reorder': 'w',
                'reset_order': 'w',
                'get_siblings': 'r'
            }.get(request.action, 'w')
        }

    def get_model_name(self, request):
        return request.path_kwargs['model']

    def get_model(self, request):
        model_name = self.get_model_name(request)
        MappedBase = get_mapped_base(request)

        if model_name not in MappedBase.classes:
            raise NotFound

        return MappedBase.classes[model_name]

    def get_model_lookup_field(self, request):
        mapper = inspect_uniform(self.get_model(request))
        return mapper.primary_key[0].name

    def get_serializer_class(self, request):
        Model = self.get_model(request)
        return get_model_serializer(Model)

    def get_filter_class(self, request):
        Model = self.get_model(request)
        return get_model_filter_class(request, Model)

    def get_queryset(self, request):
        Model = self.get_model(request)
        queryset = request.session.query(Model)

        mapper = inspect_uniform(Model)
        auto_pk = getattr(mapper.tables[0], '__jet_auto_pk__', False) if len(mapper.tables) else None
        if auto_pk:
            queryset = queryset.filter(mapper.primary_key[0].isnot(None))

        return queryset

    def filter_queryset(self, request, queryset):
        queryset = super(ModelViewSet, self).filter_queryset(request, queryset)
        if request.action == 'list':
            Model = self.get_model(request)
            queryset = apply_default_ordering(Model, queryset)
        return queryset

    @action(methods=['post'], detail=False)
    def bulk_create(self, request, *args, **kwargs):
        if not isinstance(request.data, list):
            return JSONResponse({'error': 'Request body should be an array'}, status=status.HTTP_400_BAD_REQUEST)

        self.apply_timezone(request)

        result = []

        for item in request.data:
            serializer = self.get_serializer(request, data=item)

            try:
                serializer.is_valid(raise_exception=True)
                self.perform_create(request, serializer)
                result.append({'success': True})
            except ValidationError as e:
                result.append({'success': False, 'errors': serialize_validation_error(e)})
            except Exception as e:
                result.append({'success': False, 'errors': {'non_field_errors': str(e)}})

        return JSONResponse(result, status=status.HTTP_200_OK)

    @action(methods=['get'], detail=False)
    def aggregate(self, request, *args, **kwargs):
        self.apply_timezone(request)
        queryset = self.filter_queryset(request, self.get_queryset(request))

        lookup_field = self.get_model_lookup_field(request)
        y_func = request.get_argument('_y_func').lower()
        y_column = request.get_argument('_y_column', lookup_field)

        model_name = self.get_model_name(request)
        model_serializer = self.get_serializer(request)

        if y_func == 'count':
            y_serializer = fields.IntegerField()
        else:
            y_serializers = list(filter(lambda x: x.field_name == y_column, model_serializer.fields))

            if len(y_serializers) == 0:
                raise ValidationError('Table "{}" does not have column "{}"'.format(model_name, y_column))

            y_serializer = y_serializers[0]

        filter_instance = ModelAggregateFilter()
        filter_instance.model = self.get_model(request)

        try:
            data = filter_instance.filter(queryset, {
                'y_func': y_func,
                'y_column': y_column
            })
        except SQLAlchemyError:
            queryset.session.rollback()
            raise

        result = y_serializer.to_representation(data)  # TODO: Refactor serializer

        return JSONResponse({
            'y_func': result
        })

    @action(methods=['get'], detail=False)
    def group(self, request, *args, **kwargs):
        self.apply_timezone(request)
        queryset = self.filter_queryset(request, self.get_queryset(request))

        lookup_field = self.get_model_lookup_field(request)
        x_columns = request.get_arguments('_x_column')
        x_lookup_names = request.get_arguments('_x_lookup', None)
        y_func = request.get_argument('_y_func').lower()
        y_column = request.get_argument('_y_column', lookup_field)

        # model_serializer = self.get_serializer(request)

        # x_serializers = list(filter(lambda x: x.field_name == x_column, model_serializer.fields))
        # x_serializer = x_serializers[0]

        # y_serializers = list(filter(lambda x: x.field_name == y_column, model_serializer.fields))
        # y_serializer = y_serializers[0]

        filter_instance = ModelGroupFilter()
        filter_instance.model = self.get_model(request)
        queryset = filter_instance.filter(queryset, {
            'x_columns': x_columns,
            'x_lookups': x_lookup_names,
            'y_func': y_func,
            'y_column': y_column
        })

        def map_item(row):
            if isinstance(row, Row):
                row = dict(row)

            if isinstance(row, dict):
                if 'group_1' in row:
                    row['group'] = row['group_1']
                    del row['group_1']

                return row
            else:
                return row

        try:
            instance = list(map(lambda x: map_item(x), queryset))
        except SQLAlchemyError:
            queryset.session.rollback()
            raise

        return JSONResponse(instance)
        # serializer = ModelGroupSerializer(
        #     instance=instance,
        #     many=True,
        #     # TODO: Refactor serializer
        #     # group_serializer=x_serializer,
        #     # y_func_serializer=y_serializer
        # )
        #
        # return JSONResponse(serializer.representation_data)

    @action(methods=['post'], detail=False)
    def reorder(self, request, *args, **kwargs):
        self.apply_timezone(request)
        queryset = self.filter_queryset(request, self.get_queryset(request))
        Model = self.get_model(request)
        ReorderSerializer = get_reorder_serializer(Model, queryset, request.session)

        serializer = ReorderSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        return JSONResponse(serializer.representation_data)

    @action(methods=['post'], detail=False)
    def reset_order(self, request, *args, **kwargs):
        self.apply_timezone(request)
        queryset = self.filter_queryset(request, self.get_queryset(request))
        Model = self.get_model(request)
        ResetOrderSerializer = get_reset_order_serializer(Model, queryset, request.session)

        serializer = ResetOrderSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        return JSONResponse(serializer.representation_data)

    @action(methods=['get'], detail=True)
    def get_siblings(self, request, *args, **kwargs):
        self.apply_timezone(request)
        queryset = self.filter_queryset(request, self.get_queryset(request))
        obj = self.get_object(request)
        Model = self.get_model(request)
        result = get_model_siblings(request, Model, obj, queryset)

        return JSONResponse(result)



import base64
import gzip


def decompress_data(value):
    try:
        bytes = base64.b64decode(value)
        data = gzip.decompress(bytes)
        return data.decode('utf-8')
    except AttributeError:
        return value.decode('zlib')


def compress_data(data):
    try:
        encoded = data.encode('utf-8')
        bytes = gzip.compress(encoded)
        return str(base64.b64encode(bytes), 'utf-8')
    except AttributeError:
        return data.encode('zlib')



from jet_bridge_base import status
from jet_bridge_base.configuration import configuration
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.utils.track_database import track_database_async
from jet_bridge_base.utils.track_model import track_model_async


class CreateAPIViewMixin(object):

    def create(self, request, *args, **kwargs):
        track_database_async(request)

        self.apply_timezone(request)
        serializer = self.get_serializer(request, data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(request, serializer)

        representation_data = serializer.representation_data
        track_model_async(request, kwargs.get('model'), 'create', None, representation_data)

        return JSONResponse(representation_data, status=status.HTTP_201_CREATED)

    def perform_create(self, request, serializer):
        serializer_instance = serializer.create_instance(serializer.validated_data)
        configuration.on_model_pre_create(request.path_kwargs['model'], serializer_instance)
        instance = serializer.save()
        configuration.on_model_post_create(request.path_kwargs['model'], instance)



from sqlalchemy.exc import SQLAlchemyError

from jet_bridge_base.utils.timezones import FixedOffsetTimezone

from .common import sql_get_session_engine


def sql_fetch_postgresql_default_timezone(session):
    try:
        cursor = session.execute('SELECT NOW()')
        row = cursor.fetchone()
        session.commit()

        return row[0].tzinfo
    except SQLAlchemyError:
        session.rollback()


def sql_fetch_mysql_default_timezone(session):
    try:
        cursor = session.execute('SELECT TIMEDIFF(NOW(), UTC_TIMESTAMP)')
        row = cursor.fetchone()
        session.commit()

        return FixedOffsetTimezone(row[0])
    except SQLAlchemyError:
        session.rollback()


def sql_fetch_mssql_default_timezone(session):
    try:
        cursor = session.execute('SELECT sysdatetimeoffset() as now')
        row = cursor.fetchone()
        session.commit()

        return row[0].tzinfo
    except SQLAlchemyError:
        session.rollback()


def sql_fetch_default_timezone(session):
    try:
        if sql_get_session_engine(session) == 'postgresql':
            return sql_fetch_postgresql_default_timezone(session)
        elif sql_get_session_engine(session) == 'mssql':
            return sql_fetch_mssql_default_timezone(session)
        elif sql_get_session_engine(session) == 'mysql':
            return sql_fetch_mysql_default_timezone(session)
    except:
        pass



import io
from six.moves import urllib
from PIL import Image

from jet_bridge_base.configuration import configuration
from jet_bridge_base.exceptions.not_found import NotFound
from jet_bridge_base.media_cache import cache
from jet_bridge_base.responses.redirect import RedirectResponse
from jet_bridge_base.views.base.api import APIView


class ImageResizeView(APIView):

    def create_thumbnail(self, file, thumbnail_path, max_width, max_height):
        img = Image.open(file)
        img.thumbnail((max_width, max_height), Image.ANTIALIAS)

        with io.BytesIO() as memory_file:
            img.save(memory_file, format=img.format, quality=85)  # TODO: determine real extension from format
            memory_file.seek(0)
            configuration.media_save(thumbnail_path, memory_file.read())

    def get(self, request, *args, **kwargs):
        # TODO: Move to serializer
        # TODO: Add options dependant cache name

        path = request.get_argument('path')
        max_width = request.get_argument('max_width', 320)
        max_height = request.get_argument('max_height', 240)
        external_path = path.startswith('http://') or path.startswith('https://')

        try:
            if not cache.exists(path):
                thumbnail_full_path = cache.full_path(path)

                if not external_path:
                    if not configuration.media_exists(path):
                        raise NotFound

                    file = configuration.media_open(path)
                else:
                    fd = urllib.request.urlopen(path)
                    file = io.BytesIO(fd.read())

                with file:
                    cache.clear_cache_if_needed()
                    self.create_thumbnail(file, thumbnail_full_path, max_width, max_height)
                    cache.add_file(path)

            # self.set_header('Content-Type', 'image/jpg')
            #
            # with open(thumbnail_full_path, 'rb') as f:
            #     data = f.read()
            #     self.write(data)
            # self.finish()

            # self.redirect(cache.url(path))
            return RedirectResponse(cache.url(path, request))
        except IOError as e:
            raise e



import json

from jet_bridge_base.db_types import get_session_engine, MongoColumn
from jet_bridge_base.serializers.model_serializer import get_column_data_type
from jet_bridge_base.utils.classes import is_instance_or_subclass
from sqlalchemy import Unicode, and_, or_
from sqlalchemy.dialects.postgresql import ENUM, JSONB, array
from sqlalchemy.sql import sqltypes
from six import string_types

from jet_bridge_base.fields import field, BooleanField, CharField
from jet_bridge_base.filters import lookups

EMPTY_VALUES = ([], (), {}, '', None)


def safe_is_float(value):
    try:
        float(value)
        return True
    except (ValueError, TypeError):
        return False


def safe_equals(queryset, column, value):
    if isinstance(column, MongoColumn):
        return column.__eq__(value)
    else:
        field_type = column.property.columns[0].type if hasattr(column, 'property') else column.type

        if is_instance_or_subclass(field_type, (sqltypes.JSON,)):
            if get_session_engine(queryset.session) == 'postgresql':
                value_str = json.dumps(str(value))
                operators = [column.cast(JSONB).op('@>')(value_str)]

                if safe_is_float(value):
                    operators.append(column.cast(JSONB).op('@>')(value))

                return or_(*operators)
            else:
                return column.cast(Unicode).ilike('%{}%'.format(value))
        else:
            return column.__eq__(value)


def safe_in(queryset, column, value):
    if isinstance(column, MongoColumn):
        return column.in_(value)
    else:
        field_type = column.property.columns[0].type if hasattr(column, 'property') else column.type

        if is_instance_or_subclass(field_type, (sqltypes.JSON,)):
            if get_session_engine(queryset.session) == 'postgresql':
                operators = []

                for value_item in value:
                    value_item_str = json.dumps(value_item)
                    operators.append(column.cast(JSONB).op('@>')(value_item_str))

                    if safe_is_float(value_item):
                        operators.append(column.cast(JSONB).op('@>')(value_item))

                return or_(*operators)
            else:
                operators = list(map(lambda x: column.cast(Unicode).ilike('%{}%'.format(x)), value))
                return or_(*operators)
        else:
            return column.in_(value)


def safe_startswith(queryset, column, value):
    if isinstance(column, MongoColumn):
        return column.ilike('{}%'.format(value))
    else:
        field_type = column.property.columns[0].type if hasattr(column, 'property') else column.type

        if is_instance_or_subclass(field_type, (ENUM, sqltypes.NullType)):
            return column.cast(Unicode).ilike('{}%'.format(value))
        else:
            return column.ilike('{}%'.format(value))


def safe_endswith(queryset, column, value):
    if isinstance(column, MongoColumn):
        return column.ilike('%{}'.format(value))
    else:
        field_type = column.property.columns[0].type if hasattr(column, 'property') else column.type

        if is_instance_or_subclass(field_type, (ENUM, sqltypes.NullType)):
            return column.cast(Unicode).ilike('%{}'.format(value))
        else:
            return column.ilike('%{}'.format(value))


def safe_icontains(queryset, column, value):
    if isinstance(column, MongoColumn):
        return column.ilike('%{}%'.format(value))
    else:
        field_type = column.property.columns[0].type if hasattr(column, 'property') else column.type

        if is_instance_or_subclass(field_type, (ENUM, sqltypes.NullType)):
            return column.cast(Unicode).ilike('%{}%'.format(value))
        else:
            return column.ilike('%{}%'.format(value))


def json_icontains(queryset, column, value):
    if isinstance(column, MongoColumn):
        return column.json_icontains(value)
    else:
        field_type = column.property.columns[0].type if hasattr(column, 'property') else column.type

        if is_instance_or_subclass(field_type, (sqltypes.JSON, sqltypes.NullType)) or not hasattr(column, 'astext'):
            return column.cast(Unicode).ilike('%{}%'.format(value))
        else:
            return column.astext.ilike('%{}%'.format(value))


def is_null(queryset, column, value):
    if value:
        return column.__eq__(None)
    else:
        return column.isnot(None)


def is_empty(queryset, column, value):
    field_type = column.property.columns[0].type if hasattr(column, 'property') else column.type

    if is_instance_or_subclass(field_type, sqltypes.String):
        if value:
            return or_(column.__eq__(None), column == '')
        else:
            return and_(column.isnot(None), column != '')
    else:
        return is_null(queryset, column, value)


def coveredby(queryset, column, value):
    return column.ST_CoveredBy(value)


def safe_not_array(value):
    if isinstance(value, list):
        if len(value):
            return value[0]
        else:
            return ''
    else:
        return value


def safe_array(value):
    if isinstance(value, list):
        return value
    elif isinstance(value, string_types):
        if value != '':
            return value.split(',')
        else:
            return []
    else:
        return [value]


class Filter(object):
    field_class = field
    lookup_operators = {
        lookups.EXACT: {'operator': False, 'func': safe_equals, 'pre_process': lambda x: safe_not_array(x)},
        lookups.GT: {'operator': '__gt__', 'pre_process': lambda x: safe_not_array(x)},
        lookups.GTE: {'operator': '__ge__', 'pre_process': lambda x: safe_not_array(x)},
        lookups.LT: {'operator': '__lt__', 'pre_process': lambda x: safe_not_array(x)},
        lookups.LTE: {'operator': '__le__', 'pre_process': lambda x: safe_not_array(x)},
        lookups.ICONTAINS: {'operator': False, 'func': safe_icontains, 'pre_process': lambda x: safe_not_array(x)},
        lookups.IN: {'operator': False, 'func': safe_in, 'field_kwargs': {'many': True}, 'pre_process': lambda x: safe_array(x)},
        lookups.STARTS_WITH: {'operator': False, 'func': safe_startswith, 'pre_process': lambda x: safe_not_array(x)},
        lookups.ENDS_WITH: {'operator': False, 'func': safe_endswith, 'pre_process': lambda x: safe_not_array(x)},
        lookups.IS_NULL: {'operator': False, 'func': is_null, 'field_class': BooleanField, 'pre_process': lambda x: safe_not_array(x)},
        lookups.IS_EMPTY: {'operator': False, 'func': is_empty, 'field_class': BooleanField, 'pre_process': lambda x: safe_not_array(x)},
        lookups.JSON_ICONTAINS: {'operator': False, 'func': json_icontains, 'field_class': CharField, 'pre_process': lambda x: safe_not_array(x)},
        lookups.COVEREDBY: {'operator': False, 'func': coveredby, 'pre_process': lambda x: safe_not_array(x)}
    }

    def __init__(self, name=None, column=None, lookup=lookups.DEFAULT_LOOKUP, exclude=False):
        self.name = name
        self.column = column
        self.lookup = lookup
        self.exclude = exclude

    def clean_value(self, value):
        return value

    def get_default_lookup_field_class(self):
        return get_column_data_type(self.column)

    def get_lookup_criterion(self, qs, value):
        lookup_operator = self.lookup_operators[self.lookup]
        operator = lookup_operator['operator']
        pre_process = lookup_operator.get('pre_process')
        post_process = lookup_operator.get('post_process')
        field_class = lookup_operator.get('field_class')
        field_kwargs = lookup_operator.get('field_kwargs', {})
        func = lookup_operator.get('func')

        if pre_process:
            value = pre_process(value)

        if not field_class:
            field_class = self.get_default_lookup_field_class()

        if value is not None:
            value = field_class(**field_kwargs).to_internal_value(value)

        if post_process:
            value = post_process(value)

        if func:
            if self.exclude:
                return ~func(qs, self.column, value)
            else:
                return func(qs, self.column, value)
        elif callable(operator):
            op = operator(value)

            if self.exclude:
                return ~getattr(self.column, op[0])(op[1])
            else:
                return getattr(self.column, op[0])(op[1])
        else:
            if self.exclude:
                return ~getattr(self.column, operator)(value)
            else:
                return getattr(self.column, operator)(value)

    def apply_lookup(self, qs, value):
        criterion = self.get_lookup_criterion(qs, value)
        return qs.filter(criterion)

    def filter(self, qs, value):
        value = self.clean_value(value)
        if value in EMPTY_VALUES:
            return qs
        return self.apply_lookup(qs, value)



from jet_bridge_base.db_types import empty_filter, inspect_uniform
from jet_bridge_base.filters.char_filter import CharFilter
from jet_bridge_base.filters.filter import EMPTY_VALUES
from jet_bridge_base.serializers.sql import SqlSerializer


def get_model_segment_filter(request, Model):
    mapper = inspect_uniform(Model)
    primary_key = mapper.primary_key[0].name

    class ModelSegmentFilter(CharFilter):
        def filter(self, qs, value):
            value = self.clean_value(value)
            if value in EMPTY_VALUES:
                return qs

            body = self.handler.data

            if not isinstance(body, dict):
                return qs.filter(empty_filter(Model))

            items = list(filter(lambda x: x.get('name') == value, body.get('segments', [])))

            if len(items) == 0:
                return qs.filter(empty_filter(Model))

            query = items[0].get('query')

            serializer = SqlSerializer(data={'query': query}, context={'request': request})
            serializer.is_valid(raise_exception=True)
            result = serializer.execute()
            columns = list(result['columns'])
            rows = result['data']

            if len(columns) == 0 or len(rows) == 0:
                return qs.filter(empty_filter(Model))

            ids = list(map(lambda x: list(x)[0], rows))

            return qs.filter(getattr(Model, primary_key).in_(ids))

    return ModelSegmentFilter



from jet_bridge_base.configuration import configuration
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.store import store
from jet_bridge_base.views.base.api import BaseAPIView


class ApiView(BaseAPIView):

    def get(self, request, *args, **kwargs):
        return JSONResponse({
            'version': configuration.get_version(),
            'type': configuration.get_type(),
            'store_available': store.is_ok(),
            'media_url_template': configuration.media_url('{}', request)
        })



from .common import mongo_inspect
from .mongo_base import MongoBase, MongoBaseClasses
from .mongo_column import MongoColumn
from .mongo_declarative_meta import MongoDeclarativeMeta
from .mongo_desc import MongoDesc
from .mongo_mapper import MongoMapper
from .mongo_metadata import MongoMetadata
from .mongo_metadata_file import mongo_dump_metadata_file, mongo_load_metadata_file
from .mongo_operator import MongoOperator
from .mongo_queryset import MongoQueryset
from .mongo_record import MongoRecordMeta, MongoRecord
from .mongo_session import MongoSession
from .mongo_table import MongoTable
from .mongo_reflect import reflect_mongodb
from .mongo_db import mongodb_init_database_connection, mongo_load_mapped_base



from sqlalchemy.orm import DeclarativeMeta, aliased

from .mongo import MongoDeclarativeMeta, MongoSession, mongo_inspect
from .sql import sql_inspect, sql_get_session_engine


def inspect_uniform(cls):
    if isinstance(cls, DeclarativeMeta):
        return sql_inspect(cls)
    elif isinstance(cls, MongoDeclarativeMeta):
        return mongo_inspect(cls)


def aliased_uniform(cls):
    if isinstance(cls, DeclarativeMeta):
        return aliased(cls)
    elif isinstance(cls, MongoDeclarativeMeta):
        return cls


def get_session_engine(session):
    if isinstance(session, MongoSession):
        return 'mongo'
    else:
        return sql_get_session_engine(session)



import logging

logger = logging.getLogger('jet_bridge')
ch = logging.StreamHandler()


class Formatter(logging.Formatter):

    formats = {
        logging.INFO: '%(message)s'
    }
    default_format = '%(levelname)s - %(asctime)s: %(message)s'

    def formatMessage(self, record):
        return self.formats.get(record.levelno, self.default_format) % record.__dict__

formatter = Formatter('%(asctime)s %(levelname)s %(message)s', '%Y-%m-%d %H:%M:%S')

ch.setFormatter(formatter)

logger.propagate = False
logger.addHandler(ch)

def set_logger_level(level):
    ch.setLevel(level)
    logger.setLevel(level)



from sqlalchemy.ext import automap
from sqlalchemy.ext.automap import automap_base


def _is_many_to_many(automap_base, table):
    return None, None, None


# Monkey patch to make M2M reflect as a normal Entity class
setattr(automap, '_is_many_to_many', _is_many_to_many)



from jet_bridge_base.serializers.relationship_override import ModelDescriptionRelationOverridesSerializer

from jet_bridge_base.permissions import HasProjectPermissions
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.views.base.api import APIView


class ModelDescriptionRelationshipOverrideView(APIView):
    permission_classes = (HasProjectPermissions,)

    def post(self, request, *args, **kwargs):
        serializer_context = {'request': request}
        serializer = ModelDescriptionRelationOverridesSerializer(data=request.data, many=True, context=serializer_context)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        return JSONResponse(serializer.representation_data)



import json

import requests

from jet_bridge_base import settings
from jet_bridge_base.configuration import configuration
from jet_bridge_base.encoders import JSONEncoder
from jet_bridge_base.sentry import sentry_controller


def track_model(project, environment, resource_token, model, action, uid, model_data):
    if not settings.TRACK_MODELS_ENDPOINT:
        return

    if project is None or environment is None or resource_token is None:
        error = 'MODEL_CHANGE incorrect request: {} {} {}'.format(project, environment, resource_token)
        sentry_controller.capture_message(error)
        return

    url = '{}/model_change'.format(settings.TRACK_MODELS_ENDPOINT)
    headers = {
        'Content-Type': 'application/json'
    }
    data = {
        'project': project,
        'environment': environment,
        'resource_token': resource_token,
        'model': model,
        'action': action,
        'data': model_data
    }

    if uid is not None:
        data['id'] = uid

    if settings.TRACK_MODELS_AUTH:
        headers['Authorization'] = settings.TRACK_MODELS_AUTH

    data_str = json.dumps(data, cls=JSONEncoder)

    try:
        r = requests.post(url, data=data_str, headers=headers)
        success = 200 <= r.status_code < 300

        if not success:
            error = 'MODEL_CHANGE request error: {} {} {}'.format(r.status_code, r.reason, r.text)
            sentry_controller.capture_message(error)
    except Exception as e:
        sentry_controller.capture_exception(e)


def track_model_async(request, model, action, uid, data):
    if not settings.TRACK_MODELS_ENDPOINT:
        return

    configuration.run_async(track_model, request.project, request.environment, request.resource_token, model, action, uid, data)



from jet_bridge_base.serializers.proxy_request import ProxyRequestSerializer
from jet_bridge_base.views.base.api import BaseAPIView


class ProxyRequestView(BaseAPIView):
    serializer_class = ProxyRequestSerializer

    def get(self, request, *args, **kwargs):
        return self.make_request(request, request.query_arguments, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.make_request(request, request.data, *args, **kwargs)

    def make_request(self, request, data, *args, **kwargs):
        serializer = ProxyRequestSerializer(data=data, context={'request': request, 'handler': self})
        serializer.is_valid(raise_exception=True)
        result = serializer.submit()
        return result



from __future__ import absolute_import, unicode_literals

import datetime
import decimal
import json  # noqa
import uuid
import six
from bson import ObjectId


class JSONEncoder(json.JSONEncoder):

    def is_aware(self, value):
        return value.utcoffset() is not None

    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            representation = obj.isoformat()
            if representation.endswith('+00:00'):
                representation = representation[:-6] + 'Z'
            return representation
        elif isinstance(obj, datetime.date):
            return obj.isoformat()
        elif isinstance(obj, datetime.time):
            if self.is_aware(obj):
                raise ValueError("JSON can't represent timezone-aware times.")
            representation = obj.isoformat()
            return representation
        elif isinstance(obj, datetime.timedelta):
            return six.text_type(obj.total_seconds())
        elif isinstance(obj, decimal.Decimal):
            return float(obj)
        elif isinstance(obj, uuid.UUID):
            return six.text_type(obj)
        elif isinstance(obj, ObjectId) or type(obj) is ObjectId:
            return six.text_type(obj)
        elif isinstance(obj, six.binary_type):
            return obj.decode('utf-8')
        elif hasattr(obj, 'tolist'):
            return obj.tolist()
        elif hasattr(obj, '__getitem__'):
            try:
                return dict(obj)
            except Exception:
                pass
        elif hasattr(obj, '__iter__'):
            return tuple(item for item in obj)
        return super(JSONEncoder, self).default(obj)



from jet_bridge_base.db import get_connection_tunnel
from jet_bridge_base.db_types import discover_connection
from jet_bridge_base.exceptions.validation_error import ValidationError
from jet_bridge_base.permissions import HasProjectPermissions
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.utils.conf import get_conf
from jet_bridge_base.views.base.api import BaseAPIView


class DiscoverConnectionView(BaseAPIView):
    permission_classes = (HasProjectPermissions,)

    def required_project_permission(self, request):
        return {
            'permission_type': 'project',
            'permission_object': 'project_customization',
            'permission_actions': ''
        }

    def get(self, request, *args, **kwargs):
        conf = get_conf(request)

        try:
            tunnel = get_connection_tunnel(conf)
            status = discover_connection(conf, tunnel)

            return JSONResponse({
                'status': status
            })
        except Exception as e:
            raise ValidationError(str(e))



import time
from datetime import date
from bson import ObjectId

from jet_bridge_base.logger import logger
from jet_bridge_base.models import data_types
from jet_bridge_base.utils.process import get_memory_usage_human

from .mongo_column import MongoColumn
from .mongo_table import MongoTable
from .mongo_metadata import MongoMetadata


def reflect_mongodb(
    cid_short,
    db,
    only=None,
    pending_connection=None,
    max_read_records=None
):
    available = db.list_collection_names()

    if only is None:
        load = available
    elif callable(only):
        load = [name for name in available if only(name)]
    else:
        load = [name for name in only]

    metadata = MongoMetadata()

    if pending_connection:
        pending_connection['tables_total'] = len(load)

    i = 0
    for name in load:
        # Wait to allow other threads execution
        time.sleep(0.01)

        logger.info('[{}] Analyzing collection "{}" ({} / {})" (Mem:{})...'.format(
            cid_short, name, i + 1, len(load), get_memory_usage_human())
        )

        page = 1
        limit = 10000
        table = MongoTable(name)

        while True:
            skip = (page - 1) * limit
            query_page = page
            items = db[name].find(skip=skip, limit=limit)
            has_items = False

            for item in items:
                has_items = True

                for key, value in item.items():
                    if value is None and key in table.columns:
                        continue

                    field_type = None
                    field_params = None

                    if value is None and key not in table.columns:
                        # field_type = None
                        pass
                    if isinstance(value, bool):
                        field_type = data_types.BOOLEAN
                    elif isinstance(value, int):
                        field_type = data_types.INTEGER
                    elif isinstance(value, str):
                        # field_type = data_types.TEXT
                        field_type = data_types.CHAR
                    elif isinstance(value, float):
                        field_type = data_types.FLOAT
                    elif isinstance(value, date):
                        field_type = data_types.DATE_TIME
                    elif isinstance(value, dict):
                        field_type = data_types.JSON
                    elif isinstance(value, list):
                        field_type = data_types.JSON
                    elif isinstance(value, ObjectId) or type(value) is ObjectId:
                        field_type = data_types.BINARY
                        field_params = {'type': 'object_id'}
                    else:
                        field_type = data_types.TEXT

                    if key in table.columns:
                        column = table.columns[key]
                    else:
                        column = MongoColumn(table, key, None)
                        table.append_column(column)

                    if column.type and column.type != field_type:
                        column.mixed_types = column.mixed_types or set()
                        column.mixed_types.add(column.type)
                        column.mixed_types.add(field_type)

                    column.type = field_type

                    if field_params:
                        column.params = field_params

            if has_items and (max_read_records is None or skip + limit < max_read_records):
                page += 1
            else:
                break

        if query_page == 1 and not has_items:
            logger.info('[{}] Collection "{}" does not have any data to analyze, skipping'.format(
                cid_short,
                name
            ))
        else:
            for column in table.columns:
                if column.type is None:
                    column.type = data_types.CHAR

                if column.mixed_types:
                    column.type = data_types.JSON

                    logger.info('[{}] Field "{}"."{}" has data stored in multiple types ({}), falling back to JSON'.format(
                        cid_short,
                        name,
                        column.name,
                        ','.join(column.mixed_types)
                    ))

            metadata.append_table(table)

        i += 1

        if pending_connection:
            pending_connection['tables_processed'] = i

    return metadata



import time

from jet_bridge_base.settings import set_settings


class Configuration(object):

    def __init__(self):
        self.init_time = time.time()

    def get_type(self):
        pass

    def get_version(self):
        pass

    def get_model_description(self, db_table):
        pass

    def get_hidden_model_description(self):
        return []

    def get_settings(self):
        pass

    def on_model_pre_create(self, model, pk):
        pass

    def on_model_post_create(self, model, instance):
        pass

    def on_model_pre_update(self, model, instance):
        pass

    def on_model_post_update(self, model, instance):
        pass

    def on_model_pre_delete(self, model, instance):
        pass

    def on_model_post_delete(self, model, instance):
        pass

    def media_get_available_name(self, path):
        pass

    def media_exists(self, path):
        pass

    def media_listdir(self, path):
        pass

    def media_get_modified_time(self, path):
        pass

    def media_open(self, path, mode='rb'):
        pass

    def media_save(self, path, content):
        pass

    def media_delete(self, path):
        pass

    def media_url(self, path, request):
        pass

    def session_set(self, request, name, value, secure=True):
        pass

    def session_get(self, request, name, default=None, decode=True, secure=True):
        pass

    def session_clear(self, request, name):
        pass

    def clean_sso_application_name(self, name):
        return name.lower().replace('-', '')

    def clean_sso_applications(self, applications):
        return dict(map(lambda x: (self.clean_sso_application_name(x[0]), x[1]), applications.items()))

    def run_async(self, func, *args, **kwargs):
        func(*args, **kwargs)


configuration = Configuration()


def set_configuration(new_configuration):
    global configuration
    configuration = new_configuration
    set_settings(configuration.get_settings())



from sqlalchemy.exc import SQLAlchemyError

from jet_bridge_base.db_types import get_queryset_limit
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.utils.track_database import track_database_async


class ListAPIViewMixin(object):

    def list(self, request, *args, **kwargs):
        track_database_async(request)

        self.apply_timezone(request)
        queryset = self.filter_queryset(request, self.get_queryset(request))

        paginate = not request.get_argument('_no_pagination', False)

        try:
            page = self.paginate_queryset(request, queryset) if paginate else None
            if page is not None:
                instance = list(page)
                serializer = self.get_serializer(request, instance=instance, many=True)
                return self.get_paginated_response(request, serializer.representation_data)
        except SQLAlchemyError:
            request.session.rollback()
            raise

        if get_queryset_limit(queryset) is None:
            queryset = queryset.limit(10000)

        try:
            instance = list(queryset)
            serializer = self.get_serializer(request, instance=instance, many=True)
            data = serializer.representation_data
            return JSONResponse(data)
        except SQLAlchemyError:
            request.session.rollback()
            raise



from jet_bridge_base import settings
from jet_bridge_base.sentry import sentry_controller
from jet_bridge_base.utils.backend import project_auth
from jet_bridge_base.utils.crypt import get_sha256_hash
from jet_bridge_base.utils.token import decompress_permissions, parse_token, JWT_TOKEN_PREFIX, USER_TOKEN_PREFIX, \
    PROJECT_TOKEN_PREFIX, BEARER_TOKEN_PREFIX, decode_jwt_token


class BasePermission(object):

    def has_permission(self, view, request):
        return True

    def has_object_permission(self, view, request, obj):
        return True


class HasProjectPermissions(BasePermission):
    def has_view_permissions(self, view_permissions, user_permissions, project_token):
        if not view_permissions:
            return True
        elif user_permissions.get('owner'):
            return True
        elif user_permissions.get('super_group'):
            return True

        if 'permissions' in user_permissions:
            permissions = decompress_permissions(user_permissions['permissions'])
        else:
            permissions = []

        view_permission_type = view_permissions.get('permission_type', '')
        view_permission_object = view_permissions.get('permission_object', '')
        view_permission_actions = view_permissions.get('permission_actions', '')

        if user_permissions.get('read_only'):
            if view_permission_type == 'model' and all(map(lambda x: x in ['r'], list(view_permission_actions))):
                return True
            elif view_permission_type == 'project' and view_permission_object in ['project_settings']:
                return True
            else:
                return False

        if not project_token:
            return False

        token_hash = get_sha256_hash(project_token.replace('-', '').lower())

        for item in permissions:
            item_type = item.get('permission_type', '')
            item_object = item.get('permission_object', '')
            item_actions = item.get('permission_actions', '')

            if view_permission_type == 'model':
                resource_token_hash = item.get('resource_token_hash', '')
                item_object_model = item_object.split('.', 1)[-1:][0]

                # TODO: make check non optional
                if resource_token_hash and resource_token_hash != token_hash:
                    continue

                if item_type != view_permission_type or item_object_model != view_permission_object:
                    continue
            else:
                if item_type != view_permission_type or item_object != view_permission_object:
                    continue

            if view_permission_actions in item_actions:
                return True

        return False

    def has_permission(self, view, request):
        token = parse_token(request.headers.get('AUTHORIZATION'))
        view_permissions = view.required_project_permission(request) if hasattr(view, 'required_project_permission') else None

        if not token:
            return False

        bridge_settings = request.get_bridge_settings()

        if bridge_settings:
            project_token = bridge_settings.get('token')
            project = bridge_settings.get('project')
        else:
            project_token = settings.TOKEN
            project = settings.PROJECT

        if token['type'] == JWT_TOKEN_PREFIX:
            result = decode_jwt_token(token['value'])

            if result is None:
                return False

            user_permissions = result.get('projects', {}).get(project)

            if user_permissions is None:
                return False

            request.project = result.get('project')
            request.environment = result.get('environment')
            request.resource_token = project_token

            user_id = result.get('user')
            if user_id is not None:
                sentry_controller.set_user({'id': user_id})
            else:
                sentry_controller.set_user(None)

            return self.has_view_permissions(view_permissions, user_permissions, project_token)
        elif token['type'] == USER_TOKEN_PREFIX:
            result = project_auth(token['value'], project_token, view_permissions, token['params'])

            # if result.get('warning'):
            #     view.headers['X-Application-Warning'] = result['warning']

            return result['result']
        elif token['type'] == PROJECT_TOKEN_PREFIX:
            result = project_auth(token['value'], project_token, view_permissions, token['params'])

            # if result.get('warning'):
            #     view.headers['X-Application-Warning'] = result['warning']

            return result['result']
        elif token['type'] == BEARER_TOKEN_PREFIX:
            return settings.BEARER_AUTH_KEY and token['value'] == settings.BEARER_AUTH_KEY
        else:
            return False


class ReadOnly(BasePermission):

    def has_permission(self, view, request):
        if not settings.READ_ONLY:
            return True
        if request.action in ['create', 'update', 'partial_update', 'destroy']:
            return False
        return True


class AdministratorPermissions(BasePermission):
    def has_permission(self, view, request):
        key = request.get_argument('key', None)

        if not key:
            return False

        if settings.BEARER_AUTH_KEY and key == settings.BEARER_AUTH_KEY:
            return True
        else:
            result = decode_jwt_token(key)

            if result is None:
                return False

            admin = result.get('admin', False)

            if not admin:
                return False

            return True



from jet_bridge_base import fields
from jet_bridge_base.fields import FloatField
from jet_bridge_base.filters.filter import Filter


class IntegerFilter(Filter):
    field_class = FloatField

    def get_default_lookup_field_class(self):
        return fields.FloatField



from social_core.backends.utils import load_backends
from social_core.utils import get_strategy


STRATEGY_PATH = 'jet_bridge_base.external_auth.strategy.JetBridgeStrategy'
STORAGE_PATH = 'jet_bridge_base.external_auth.storage.JetBridgeStorage'


def load_strategy(request_handler, request, config):
    return get_strategy(STRATEGY_PATH, STORAGE_PATH, request_handler, request, config)


def load_backends_classes(backend_paths):
    backends = load_backends(backend_paths, force_load=True)

    return dict(map(lambda x: (backend_paths[x[0]], x[1]), enumerate(backends.values())))




import os
import random
import socket
import tempfile
import threading
import time
from subprocess import Popen, PIPE

from jet_bridge_base.logger import logger


class SSHTunnel(object):
    local_bind_host = '127.0.0.1'
    local_bind_port = None
    process = None
    check_thread = None
    tunnel_timeout = 10.0

    def __init__(
        self,
        name,
        ssh_host,
        ssh_port,
        ssh_user,
        ssh_private_key,
        remote_host,
        remote_port,
        on_close=None
    ):
        self.name = name
        self.ssh_host = ssh_host
        self.ssh_port = ssh_port
        self.ssh_user = ssh_user
        self.ssh_private_key = ssh_private_key
        self.remote_host = remote_host
        self.remote_port = remote_port
        self.on_close = on_close

    def is_tunnel_alive(self):
        connect_to = (self.local_bind_host, self.local_bind_port)
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(self.tunnel_timeout)

        try:
            s.connect(connect_to)
            s.sendall('Hello, world'.encode('utf-8'))
            s.recv(1024)

            return True
        except socket.error:
            return False
        finally:
            s.close()

    def run_ssh_tunnel_process(self):
        with tempfile.NamedTemporaryFile(delete=False) as f:
            f.write(self.ssh_private_key.encode('utf-8'))
            keyfile = f.name

        listen = 'localhost:{}:{}:{}'.format(self.local_bind_port, self.remote_host, self.remote_port)
        command = ['ssh', '-N', '-L', listen, '-i', keyfile, '-o', 'StrictHostKeyChecking=no', '-o', 'ExitOnForwardFailure=yes']

        if self.ssh_port:
            command.extend(['-p', str(self.ssh_port)])

        command.extend(['{}@{}'.format(self.ssh_user, self.ssh_host)])

        logger.debug('Running SSH tunnel: {}'.format(command))

        try:
            process = Popen(
                command,
                stdin=PIPE,
                stdout=PIPE,
                stderr=PIPE
            )
        except FileNotFoundError:
            raise Exception('SSH is not installed')

        output = process.stdout.readlines()
        logger.debug('SSH run output: {}'.format(output))

        os.unlink(keyfile)

        return_code = process.poll()

        if return_code is not None:
            error_output = process.stderr.readlines()
            logger.debug('SSH terminated with error: {}'.format(error_output))

            error = '\n'.join(map(lambda x: x.decode('utf-8'), error_output))
            raise Exception(error)

        return process

    def execute_check_thread(self, process):
        while True:
            time.sleep(5)
            return_code = process.poll()

            if return_code is not None:
                logger.info('SSH tunnel is terminated (CODE:{})'.format(return_code))
                break
            elif not self.is_tunnel_alive():
                logger.info('SSH tunnel is dropped')
                process.kill()
                break

        if self.on_close:
            self.on_close()

    @property
    def is_active(self):
        return self.is_tunnel_alive()

    def is_port_used(self, port):
        connect_to = (self.local_bind_host, port)
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        try:
            if s.connect_ex(connect_to) == 0:
                return True
            else:
                return False
        finally:
            s.close()

    def get_unused_port(self):
        while True:
            port = random.randint(10000, 65535)
            if not self.is_port_used(port):
                return port

    def start(self):
        self.local_bind_port = self.get_unused_port()
        self.process = self.run_ssh_tunnel_process()

        self.check_thread = threading.Thread(
            target=self.execute_check_thread,
            args=(self.process,),
            name='Tunnel-check-{}'.format(self.name)
        )
        self.check_thread.start()

    def close(self):
        if self.process:
            self.process.kill()



from jet_bridge_base import fields
from jet_bridge_base.exceptions.validation_error import ValidationError
from jet_bridge_base.serializers.serializer import Serializer


class TableColumnParamsRelatedModelSerializer(Serializer):
    model = fields.CharField()


class TableColumnParamsSerializer(Serializer):
    length = fields.IntegerField(required=False)
    related_model = TableColumnParamsRelatedModelSerializer(required=False)
    custom_primary_key = fields.CharField(required=False)


class TableColumnSerializer(Serializer):
    name = fields.CharField()
    field = fields.CharField()
    db_field = fields.CharField(required=False)
    primary_key = fields.BooleanField(required=False)
    null = fields.BooleanField(required=False)
    default_type = fields.CharField(required=False)
    default_value = fields.RawField(required=False)
    params = TableColumnParamsSerializer(required=False)
    data_source_field = fields.CharField(required=False)
    data_source_name = fields.CharField(required=False)
    data_source_params = fields.JSONField(required=False)


class TableSerializer(Serializer):
    name = fields.CharField()
    columns = TableColumnSerializer(many=True)
    data_source_name = fields.CharField(required=False)
    data_source_name_plural = fields.CharField(required=False)

    def validate(self, attrs):
        if not any(map(lambda x: x.get('primary_key'), attrs['columns'])):
            raise ValidationError('No primary key specified')
        return attrs



from .common import sql_inspect, sql_get_session_engine
from .sql_reflect import sql_get_tables, sql_reflect
from .sql_db import sql_init_database_connection, sql_build_engine_url, sql_create_connection_engine, sql_load_mapped_base, sql_load_database_table
from .sql_metadata_file import sql_dump_metadata_file, sql_load_metadata_file
from .timezones import sql_fetch_default_timezone



from jet_bridge_base import fields


class SqlParamsSerializers(fields.CharField):

    def to_internal_value_item(self, value):
        value = super(SqlParamsSerializers, self).to_internal_value_item(value)
        if value is None:
            return []
        # value = list(filter(lambda x: x != '', value.split(',')))
        value = value.split(',')
        return dict([['param_{}'.format(i), x] for i, x in enumerate(value)])

    def to_representation_item(self, value):
        return list(value)



from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from jet_bridge_base import settings
from jet_bridge_base.logger import logger
from jet_bridge_base.models.base import Base


class Store(object):
    engine = None
    sessions = None

    def __init__(self, path, base):
        self.path = path
        self.base = base
        self.open()

    def open(self):
        try:
            self.engine = create_engine('sqlite:///{}'.format(self.path))
            self.sessions = scoped_session(sessionmaker(self.engine))

            import jet_bridge_base.models
            Base.metadata.create_all(self.engine)
        except Exception as e:
            logger.error('Store initialize failed', exc_info=e)

    def close(self):
        if self.engine:
            self.engine.dispose()

    def is_ok(self):
        return self.engine is not None and self.sessions is not None

    def session(self):
        return self.sessions()


store = Store(settings.STORE_PATH, Base)



from jet_bridge_base.db_types import empty_filter, inspect_uniform
from jet_bridge_base.filters.char_filter import CharFilter
from jet_bridge_base.filters.filter import EMPTY_VALUES


def get_model_m2m_filter(Model):
    mapper = inspect_uniform(Model)

    class ModelM2MFilter(CharFilter):

        def filter(self, qs, value):
            if value in EMPTY_VALUES:
                return qs

            params = value.split(',', 2)

            if len(params) < 2:
                return qs.filter(empty_filter(Model))

            relation_name, value = params

            relations = []

            for relationship in mapper.relationships:
                for sub_relationship in relationship.mapper.relationships:
                    if sub_relationship.table.name != relation_name:
                        continue

                    relations.append({
                        'relationship': relationship,
                        'sub_relationship': sub_relationship
                    })

            if len(relations) == 0:
                return qs.filter(empty_filter(Model))

            relation = relations[0]
            relationship_entity = relation['relationship'].mapper.entity
            id_column_name = relation['sub_relationship'].primaryjoin.right.name
            relationship_entity_id_key = getattr(relationship_entity, id_column_name)

            qs = qs.join(relationship_entity)
            qs = qs.filter(relationship_entity_id_key == value)

            return qs

    return ModelM2MFilter



import time

from sqlalchemy.sql.base import _bind_or_error
from sqlalchemy import Table
from sqlalchemy import exc
from sqlalchemy import inspection
from sqlalchemy import util

from jet_bridge_base import settings
from jet_bridge_base.logger import logger
from jet_bridge_base.utils.process import get_memory_usage_human


def sql_get_tables(
    insp,
    metadata,
    bind=None,
    schema=None,
    foreign=False,
    views=False,
    only=None,
    extend_existing=False
):
    view_names = []

    available = util.OrderedSet(insp.get_table_names(schema))

    if foreign and hasattr(insp, 'get_foreign_table_names'):
        table_names = insp.get_foreign_table_names()
        available.update(table_names)

    if views:
        try:
            view_names = insp.get_view_names(schema)
            available.update(view_names)
        except NotImplementedError:
            pass

    if schema is not None:
        available_w_schema = util.OrderedSet(
            ["%s.%s" % (schema, name) for name in available]
        )
    else:
        available_w_schema = available

    current = set(metadata.tables)

    if only is None:
        load = [
            name
            for name, schname in zip(available, available_w_schema)
            if extend_existing or schname not in current
        ]
    elif callable(only):
        load = [
            name
            for name, schname in zip(available, available_w_schema)
            if (extend_existing or schname not in current) and only(name)
        ]
    else:
        missing = [name for name in only if name not in available]
        if missing:
            s = schema and (" schema '%s'" % schema) or ""
            raise exc.InvalidRequestError(
                "Could not reflect: requested table(s) not available "
                "in %r%s: (%s)" % (bind.engine, s, ", ".join(missing))
            )
        load = [
            name
            for name in only
            if extend_existing or name not in current
        ]

    return load, view_names


def sql_reflect(
    cid_short,
    metadata,
    bind=None,
    schema=None,
    foreign=False,
    views=False,
    only=None,
    extend_existing=False,
    autoload_replace=True,
    resolve_fks=True,
    pending_connection=None,
    **dialect_kwargs
):
    if bind is None:
        bind = _bind_or_error(metadata)

    with inspection.inspect(bind)._inspection_context() as insp:
        reflect_opts = {
            "autoload_with": insp,
            "extend_existing": extend_existing,
            "autoload_replace": autoload_replace,
            "resolve_fks": resolve_fks,
            "_extend_on": set(),
        }

        reflect_opts.update(dialect_kwargs)

        if schema is None:
            schema = metadata.schema

        if schema is not None:
            reflect_opts["schema"] = schema

        load, view_names = sql_get_tables(insp, metadata, bind, schema, foreign, views, only, extend_existing)

        """
        Modified: Added default PK set and progress display
        """

        if pending_connection:
            pending_connection['tables_total'] = len(load)

        i = 0
        for name in load:
            if name in ['pg_stat_statements']:
                continue

            # Wait to allow other threads execution
            time.sleep(0.01)

            try:
                logger.info('[{}] Analyzing table "{}" ({} / {})" (Mem:{})...'.format(cid_short, name, i + 1, len(load), get_memory_usage_human()))
                table = Table(name, metadata, **reflect_opts)

                if bind.engine.name == 'clickhouse':
                    table = metadata.tables[table.key]

                if view_names and name in view_names:
                    setattr(table, '__jet_is_view__', True)

                args = []
                has_pk = False
                first_column = None

                for item in table.columns:
                    if not has_pk and item.primary_key:
                        has_pk = True
                    if first_column is None:
                        first_column = item

                if not has_pk and first_column is not None:
                    logger.warning('[{}] Table "{}" is missing PK: "{}" column was set as PK'.format(cid_short, name, first_column.name))

                    first_column.primary_key = True
                    args.append(first_column)
                    reflect_opts['extend_existing'] = True
                    table = Table(name, metadata, *args, **reflect_opts)
                    setattr(table, '__jet_auto_pk__', True)


            except exc.UnreflectableTableError as uerr:
                util.warn("Skipping table %s: %s" % (name, uerr))

            i += 1

            if settings.DATABASE_MAX_TABLES is not None and i >= settings.DATABASE_MAX_TABLES:
                logger.warning('[{}] Max tables limit ({}) reached'.format(cid_short, settings.DATABASE_MAX_TABLES))
                break

            if pending_connection:
                pending_connection['tables_processed'] = i

    """
    Modify END
    """



import dateparser
from datetime import datetime

from .common import get_session_engine
from .mongo import MongoSession
from .sql import sql_fetch_default_timezone


def fetch_default_timezone(session):
    if isinstance(session, MongoSession):
        pass
    else:
        return sql_fetch_default_timezone(session)


def apply_session_timezone(session, timezone):
    if isinstance(session, MongoSession):
        pass
    else:
        if get_session_engine(session) == 'mysql':
            session.execute('SET time_zone = :tz', {'tz': timezone})
            session.info['_queries_timezone'] = timezone
        elif get_session_engine(session) in ['postgresql', 'mssql']:
            offset_hours = dateparser.parse(datetime.now().isoformat() + timezone).utcoffset().total_seconds() / 60 / 60
            offset_hours_str = '{:+}'.format(offset_hours).replace(".0", "")
            session.execute('SET TIME ZONE :tz', {'tz': offset_hours_str})
            session.info['_queries_timezone'] = timezone



import time
from collections import OrderedDict
import math

from jet_bridge_base.db_types import queryset_count_optimized
from jet_bridge_base.exceptions.missing_argument_error import MissingArgumentError
from jet_bridge_base.paginators.pagination import Pagination
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.utils.http import replace_query_param, remove_query_param


class PageNumberPagination(Pagination):
    default_page_size = 25
    page_query_param = 'page'
    page_size_query_param = '_per_page'
    max_page_size = 10000

    count = None
    count_query_time = None
    page_number = None
    page_size = None
    data_query_time = None
    handler = None

    def paginate_queryset(self, request, queryset, handler):
        page_number = self.get_page_number(request, handler)
        if not page_number:
            return None

        page_size = self.get_page_size(request, handler)
        if not page_size:
            return None

        data_query_start = time.time()
        result = list(queryset.offset((page_number - 1) * page_size).limit(page_size))
        data_query_end = time.time()

        self.data_query_time = round(data_query_end - data_query_start, 3)

        count_query_start = time.time()
        if page_number == 1 and len(result) < page_size:
            self.count = len(result)
        else:
            self.count = queryset_count_optimized(request.session, queryset)
        count_query_end = time.time()

        self.count_query_time = round(count_query_end - count_query_start, 3)

        self.page_number = page_number
        self.page_size = page_size
        self.handler = handler

        return result

    def get_pages_count(self):
        return int(math.ceil(self.count / self.page_size)) if self.count is not None else None

    def get_paginated_response(self, request, data):
        return JSONResponse(OrderedDict([
            ('count', self.count),
            ('next', self.get_next_link(request, data)),
            ('previous', self.get_previous_link(request)),
            ('results', data),
            ('num_pages', self.get_pages_count()),
            ('per_page', self.page_size),
            ('has_more', self.has_next_potential(data)),
            ('data_query_time', self.data_query_time),
            ('count_query_time', self.count_query_time),
        ]))

    def get_page_number(self, request, handler):
        try:
            result = int(request.get_argument(self.page_query_param))
            return max(result, 1)
        except (MissingArgumentError, ValueError):
            return 1

    def get_page_size(self, request, handler):
        if self.page_size_query_param:
            try:
                result = int(request.get_argument(self.page_size_query_param))
                result = max(result, 1)

                if self.max_page_size:
                    result = min(result, self.max_page_size)

                return result
            except (MissingArgumentError, ValueError):
                pass

        return self.default_page_size

    def has_next(self):
        pages_count = self.get_pages_count()
        return self.page_number < pages_count if pages_count is not None else None

    def has_next_potential(self, data):
        has_next = self.has_next()
        if has_next is False:
            return has_next
        elif has_next is None and len(data) == 0:
            return False
        return True

    def has_previous(self):
        return self.page_number > 1

    def next_page_number(self):
        return self.page_number + 1

    def previous_page_number(self):
        return self.page_number - 1

    def get_next_link(self, request, data):
        if not self.has_next_potential(data):
            return None
        url = request.full_url()
        page_number = self.next_page_number()
        return replace_query_param(url, self.page_query_param, page_number)

    def get_previous_link(self, request):
        if not self.has_previous():
            return None
        url = request.full_url()
        page_number = self.previous_page_number()
        if page_number == 1:
            return remove_query_param(url, self.page_query_param)
        return replace_query_param(url, self.page_query_param, page_number)



import six

from jet_bridge_base import status


class APIException(Exception):
    default_detail = 'A server error occurred.'
    default_code = 'error'
    default_status_code = status.HTTP_500_INTERNAL_SERVER_ERROR

    def __init__(self, detail=None, code=None, status_code=None):
        if detail is None:
            detail = self.default_detail
        if code is None:
            code = self.default_code
        if status_code is None:
            status_code = self.default_status_code

        self.detail = detail
        self.code = code
        self.status_code = status_code

    def __str__(self):
        return six.text_type(self.detail)



import datetime
import time

from sqlalchemy import text, select, column, func, desc, or_, cast
from sqlalchemy import sql
from sqlalchemy.sql import sqltypes, quoted_name
from sqlalchemy.exc import SQLAlchemyError

from jet_bridge_base import fields
from jet_bridge_base.db import get_type_code_to_sql_type
from jet_bridge_base.db_types import get_session_engine, apply_session_timezone, get_sql_aggregate_func_by_name, \
    get_sql_group_func_lookup
from jet_bridge_base.fields.datetime import datetime_apply_default_timezone
from jet_bridge_base.exceptions.sql import SqlError
from jet_bridge_base.exceptions.validation_error import ValidationError
from jet_bridge_base.fields.sql_params import SqlParamsSerializers
from jet_bridge_base.filters import lookups
from jet_bridge_base.filters.filter import EMPTY_VALUES
from jet_bridge_base.filters.filter_for_dbfield import filter_for_data_type
from jet_bridge_base.serializers.serializer import Serializer
from jet_bridge_base.utils.db_types import map_to_sql_type, sql_to_map_type


class ColumnSerializer(Serializer):
    name = fields.CharField()
    data_type = fields.CharField()


class FilterItemSerializer(Serializer):
    name = fields.CharField()
    value = fields.RawField(allow_many=True, required=False)


class AggregateSerializer(Serializer):
    func = fields.CharField()
    column = fields.CharField(required=False)


class GroupSerializer(Serializer):
    xColumn = fields.CharField()
    xLookup = fields.CharField(required=False)
    yColumn = fields.CharField(required=False)
    yFunc = fields.CharField()


class GroupsItemSerializer(Serializer):
    xColumn = fields.CharField()
    xLookup = fields.CharField(required=False)


class GroupsSerializer(Serializer):
    xColumns = GroupsItemSerializer(many=True)
    yColumn = fields.CharField(required=False)
    yFunc = fields.CharField()


class SqlSerializer(Serializer):
    query = fields.CharField()
    offset = fields.IntegerField(required=False)
    limit = fields.IntegerField(required=False)
    order_by = fields.CharField(many=True, required=False)
    count = fields.BooleanField(default=False)
    columns = ColumnSerializer(many=True, required=False)
    filters = FilterItemSerializer(many=True, required=False)
    aggregate = AggregateSerializer(required=False)
    group = GroupSerializer(required=False)
    groups = GroupsSerializer(required=False)
    timezone = fields.CharField(required=False)
    schema = fields.CharField(required=False)
    params = SqlParamsSerializers(required=False)
    params_obj = fields.JSONField(required=False)
    v = fields.IntegerField(default=1)

    def validate(self, attrs):
        forbidden = ['insert', 'update', 'delete', 'grant', 'show']
        for i in range(len(forbidden)):
            forbidden.append('({}'.format(forbidden[i]))
        if any(map(lambda x: ' {} '.format(attrs['query'].lower()).find(' {} '.format(x)) != -1, forbidden)):
            raise ValidationError({'query': 'forbidden query'})

        if attrs['v'] < 2:
            i = 0
            while attrs['query'].find('%s') != -1:
                attrs['query'] = attrs['query'].replace('%s', ':param_{}'.format(i), 1)
                i += 1

        if 'limit' in attrs:
            if attrs['limit'] > 1000:
                attrs['limit'] = 1000
            elif attrs['limit'] < 0:
                del attrs['limit']

        return attrs

    def aggregate_queryset(self, subquery, data, session):
        func_param = data['aggregate'].get('func').lower()
        column_param = data['aggregate'].get('column')

        y_column = self.get_column(session, column_param) if column_param is not None else None
        y_func = get_sql_aggregate_func_by_name(func_param, y_column)

        if y_func is None:
            return subquery.filter(sql.false())
        else:
            return select([y_func]).select_from(subquery)

    def group_queryset(self, subquery, data, session):
        def get_y_func(group):
            y_func_param = group.get('yFunc').lower()
            y_column_param = group.get('yColumn')
            y_column = self.get_column(session, y_column_param) if y_column_param is not None else None
            return get_sql_aggregate_func_by_name(y_func_param, y_column)

        if 'groups' in data:
            y_func = get_y_func(data['groups'])
        elif 'group' in data:
            y_func = get_y_func(data['group'])
        else:
            y_func = None

        if y_func is None:
            return subquery.filter(sql.false())

        def group_name(i):
            if i == 0 and get_session_engine(session) != 'mssql':
                return 'group'
            else:
                return 'group_{}'.format(i + 1)

        def map_group_column(group, i):
            x_lookup_param = group.get('xLookup')
            x_column_param = group.get('xColumn')
            x_column = self.get_column(session, x_column_param) if x_column_param is not None else None

            lookup_params = x_lookup_param.split('_') if x_lookup_param else []
            lookup_type = lookup_params[0] if len(lookup_params) >= 1 else None
            lookup_param = lookup_params[1] if len(lookup_params) >= 2 else None

            return get_sql_group_func_lookup(session, lookup_type, lookup_param, x_column).label(group_name(i))

        if 'groups' in data:
            x_lookups = list(map(lambda x: map_group_column(x[1], x[0]), enumerate(data['groups']['xColumns'])))
        elif 'group' in data:
            x_lookups = [map_group_column(data['group'], 0)]

        x_lookup_names = list(map(lambda x: x.name, x_lookups))

        queryset = select([*x_lookups, y_func.label('y_func')]).select_from(subquery)

        if get_session_engine(session) == 'mssql':
            return queryset.group_by(*x_lookups).order_by(*x_lookup_names)
        else:
            return queryset.group_by(*x_lookup_names).order_by(*x_lookup_names)

    def filter_queryset(self, queryset, data):
        filters_instances = []
        request = self.context.get('request')
        session = request.session

        for item in data.get('columns', []):
            query_type = map_to_sql_type(item['data_type'])()
            filter_data = filter_for_data_type(query_type)
            for lookup in filter_data['lookups']:
                for exclude in [False, True]:
                    column_ = self.get_column(session, item['name'], type_=query_type)
                    instance = filter_data['filter_class'](
                        name=item['name'],
                        column=column_,
                        lookup=lookup,
                        exclude=exclude
                    )
                    filters_instances.append(instance)

        def get_filter_value(name, filters_instance=None):
            filter_items = list(filter(lambda x: x['name'] == name, data.get('filters', [])))

            if not len(filter_items):
                return

            return filter_items[0].get('value', None)

        for item in filters_instances:
            if item.name:
                argument_name = '{}__{}'.format(item.name, item.lookup)
                if item.exclude:
                    argument_name = 'exclude__{}'.format(argument_name)
                value = get_filter_value(argument_name, item)

                if value is None and item.lookup == lookups.DEFAULT_LOOKUP:
                    argument_name = item.name
                    if item.exclude:
                        argument_name = 'exclude__{}'.format(argument_name)
                    value = get_filter_value(argument_name, item)
            else:
                value = None

            queryset = item.filter(queryset, value)

        search = get_filter_value('_search')

        if search not in EMPTY_VALUES:
            def map_column(item):
                field = self.get_column(session, item['name'])
                query_type = map_to_sql_type(item['data_type'])()

                if isinstance(query_type, (sqltypes.Integer, sqltypes.Numeric)):
                    return cast(field, sqltypes.String).__eq__(search)
                elif isinstance(query_type, (sqltypes.JSON, sqltypes.ARRAY, sqltypes.Enum)):
                    return cast(field, sqltypes.String).ilike('%{}%'.format(search))
                elif isinstance(query_type, sqltypes.String):
                    return cast(field, sqltypes.String).ilike('%{}%'.format(search))

            operators = list(filter(lambda x: x is not None, map(map_column, data.get('columns', []))))
            queryset = queryset.filter(or_(*operators))

        return queryset

    def paginate_queryset(self, queryset, data):
        if 'offset' in data:
            queryset = queryset.offset(data['offset'])

        if 'limit' in data:
            if data['limit']:
                queryset = queryset.limit(data['limit'])
        elif data['v'] >= 2:
            queryset = queryset.limit(100)

        return queryset

    def map_order_field(self, session, name):
        descending = False
        if name.startswith('-'):
            name = name[1:]
            descending = True
        field = self.get_column(session, name)
        if descending:
            field = desc(field)
        return field

    def get_column(self, session, name, **kwargs):
        if get_session_engine(session) == 'oracle':
            name = quoted_name(name, True)
        return column(name, **kwargs)

    def sort_queryset(self, queryset, data, session):
        if 'order_by' in data:
            order_by = list(map(lambda x: self.map_order_field(session, x), data['order_by']))
            queryset = queryset.order_by(*order_by)
        else:
            if 'aggregate' not in data and 'group' not in data and 'groups' not in data:
                if get_session_engine(session) == 'mssql':
                    for item in data.get('columns', []):
                        field = self.get_column(session, item['name'])
                        queryset = queryset.order_by(field)
                        break

        return queryset

    def execute(self, data):
        request = self.context.get('request')
        session = request.session

        query = data['query']

        if data['v'] >= 2:
            params = data.get('params_obj', {})
        else:
            params = data.get('params', [])

        if data.get('timezone') is not None:
            try:
                apply_session_timezone(session, data['timezone'])
            except SQLAlchemyError:
                session.rollback()
                pass

        if 'schema' in data:
            try:
                session.execute('SET search_path TO :schema', {'schema': data['schema']})
            except SQLAlchemyError:
                session.rollback()
                pass

        subquery = text(query).columns().subquery('__jet_q2')
        count_rows = None
        count_query_time = None

        if data['count']:
            try:
                count_queryset = select([func.count()]).select_from(subquery)
                count_queryset = self.filter_queryset(count_queryset, data)

                count_query_start = time.time()
                count_result = session.execute(count_queryset, params)
                count_rows = count_result.all()[0][0]
                count_query_end = time.time()

                count_query_time = round(count_query_end - count_query_start, 3)
            except SQLAlchemyError:
                session.rollback()
            except Exception:
                pass

        try:
            if 'aggregate' in data:
                queryset = self.aggregate_queryset(subquery, data, session)
            elif 'groups' in data or 'group' in data:
                queryset = self.group_queryset(subquery, data, session)
            else:
                queryset = select(['*']).select_from(subquery)

            queryset = self.filter_queryset(queryset, data)

            if 'aggregate' not in data and 'group' not in data and 'groups' not in data:
                queryset = self.paginate_queryset(queryset, data)

            if 'group' not in data and 'groups' not in data:
                queryset = self.sort_queryset(queryset, data, session)

            data_query_start = time.time()
            result = session.execute(queryset, params)
            data_query_end = time.time()

            data_query_time = round(data_query_end - data_query_start, 3)

            def map_column(x):
                if x == '?column?':
                    return
                return x

            def map_row_column(x):
                if isinstance(x, bytes):
                    try:
                        return x.decode('utf-8')
                    except UnicodeDecodeError:
                        return x.hex()
                elif isinstance(x, datetime.datetime):
                    x = datetime_apply_default_timezone(x, request)
                    return x
                else:
                    return x

            def map_row(row):
                return list(map(lambda x: map_row_column(row[x]), row.keys()))

            column_names = result.keys()

            if 'groups' in data or 'group' in data:
                column_names = list(map(lambda x: 'group' if x == 'group_1' else x, column_names))

            cursor_description = result.cursor.description
            response = {
                'data': list(map(map_row, result)),
                'columns': list(map(map_column, column_names))
            }

            type_code_to_sql_type = get_type_code_to_sql_type(request)
            if type_code_to_sql_type:
                def map_column_description(column):
                    name = column.name if hasattr(column, 'name') else ''
                    sql_type = type_code_to_sql_type.get(column.type_code) if hasattr(column, 'type_code') else None
                    field = sql_to_map_type(sql_type) if sql_type else None
                    return name, {
                        'field': field
                    }

                response['column_descriptions'] = dict(map(map_column_description, cursor_description))

            if count_rows is not None:
                response['count'] = count_rows

            limit = queryset._limit
            if limit:
                response['limit'] = limit

            response['data_query_time'] = data_query_time

            if count_rows is not None:
                response['count_query_time'] = count_query_time

            return response
        except SQLAlchemyError as e:
            session.rollback()
            raise SqlError(e)
        except TypeError as e:
            raise SqlError(e)
        except Exception as e:
            raise SqlError(e)
        finally:
            session.close()


class SqlsSerializer(Serializer):
    queries = SqlSerializer(many=True)

    def execute(self, data):
        serializer = SqlSerializer(context=self.context)

        def map_query(query):
            try:
                return serializer.execute(query)
            except SqlError as e:
                return {'error': str(e.detail)}

        return list(map(map_query, data['queries']))



import six

from jet_bridge_base.fields.field import Field


class IntegerField(Field):
    field_error_messages = {
        'invalid': 'not a valid integer'
    }

    def to_internal_value_item(self, value):
        if value is None:
            return

        if value is True:
            return 1
        elif value is False:
            return 0

        value = six.text_type(value).strip()

        try:
            return int(value)
        except (ValueError, TypeError):
            self.error('invalid')

    def to_representation_item(self, value):
        if value is None:
            return
        return value



from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.utils.track_database import track_database_async


class RetrieveAPIViewMixin(object):

    def retrieve(self, request, *args, **kwargs):
        track_database_async(request)

        self.apply_timezone(request)
        instance = self.get_object(request)
        serializer = self.get_serializer(request, instance=instance)
        return JSONResponse(serializer.representation_data)



from jet_bridge_base.models import data_types
from jet_bridge_base.logger import logger
from jet_bridge_base.utils.classes import issubclass_safe, is_instance_or_subclass

from sqlalchemy.sql import sqltypes

map_data_types = [
    {'sql_type': sqltypes.VARCHAR, 'map_type': data_types.CHAR, 'db_type': data_types.CHAR},
    {'sql_type': sqltypes.CHAR, 'map_type': data_types.CHAR, 'db_type': data_types.FIXED_CHAR},
    {'sql_type': sqltypes.Unicode, 'map_type': data_types.CHAR, 'db_type': data_types.CHAR},
    {'sql_type': sqltypes.String, 'map_type': data_types.CHAR, 'db_type': data_types.CHAR},
    {'sql_type': sqltypes.Text, 'map_type': data_types.TEXT, 'db_type': data_types.TEXT},
    {'sql_type': sqltypes.Enum, 'map_type': data_types.SELECT, 'db_type': data_types.SELECT},
    {'sql_type': sqltypes.Boolean, 'map_type': data_types.BOOLEAN, 'db_type': data_types.BOOLEAN, 'convert': lambda x: '{}::boolean'.format(x)},
    {'sql_type': sqltypes.Integer, 'map_type': data_types.INTEGER, 'db_type': data_types.INTEGER, 'convert': lambda x: '{}::integer'.format(x)},
    {'sql_type': sqltypes.SmallInteger, 'map_type': data_types.INTEGER, 'db_type': data_types.SMALL_INTEGER, 'convert': lambda x: '{}::integer'.format(x)},
    {'sql_type': sqltypes.BigInteger, 'map_type': data_types.INTEGER, 'db_type': data_types.BIG_INTEGER, 'convert': lambda x: '{}::bigint'.format(x)},
    {'sql_type': sqltypes.Numeric, 'map_type': data_types.FLOAT, 'db_type': data_types.NUMBER, 'convert': lambda x: '{}::numeric'.format(x)},
    {'sql_type': sqltypes.Float, 'map_type': data_types.FLOAT, 'db_type': data_types.FLOAT, 'convert': lambda x: '{}::double precision'.format(x)},
    {'sql_type': sqltypes.DECIMAL, 'map_type': data_types.FLOAT, 'db_type': data_types.DECIMAL, 'convert': lambda x: '{}::double precision'.format(x)},
    {'sql_type': sqltypes.Date, 'map_type': data_types.DATE, 'db_type': data_types.DATE, 'convert': lambda x: '{}::text::date'.format(x)},
    {'sql_type': sqltypes.DateTime, 'map_type': data_types.DATE_TIME, 'db_type': data_types.DATE_TIME, 'convert': lambda x: '{}::timestamp with time zone'.format(x)},
    {'sql_type': sqltypes.TIMESTAMP, 'map_type': data_types.DATE_TIME, 'db_type': data_types.TIMESTAMP, 'convert': lambda x: '{}::timestamp with time zone'.format(x)},
    {'sql_type': sqltypes.JSON, 'map_type': data_types.JSON, 'db_type': data_types.JSON, 'convert': lambda x: '{}::json'.format(x)},
    {'sql_type': sqltypes.ARRAY, 'map_type': data_types.JSON, 'db_type': data_types.JSON, 'convert': lambda x: '{}::json'.format(x)},
    {'sql_type': sqltypes.BINARY, 'map_type': data_types.BINARY, 'db_type': data_types.BINARY},
    {'sql_type': sqltypes.VARBINARY, 'map_type': data_types.BINARY, 'db_type': data_types.BINARY},
]
default_sql_type = sqltypes.VARCHAR
default_map_type = data_types.CHAR
default_db_type = data_types.CHAR

try:
    from geoalchemy2 import types
    map_data_types.append({'sql_type': types.Geometry, 'map_type': data_types.GEOMETRY, 'db_type': data_types.GEOMETRY})
    map_data_types.append({'sql_type': types.Geography, 'map_type': data_types.GEOGRAPHY, 'db_type': data_types.GEOGRAPHY})
except ImportError:
    pass

try:
    from sqlalchemy.dialects import postgresql
    map_data_types.append({'sql_type': postgresql.UUID, 'map_type': data_types.UUID, 'db_type': data_types.UUID})
    map_data_types.append({'sql_type': postgresql.DOUBLE_PRECISION, 'map_type': data_types.FLOAT, 'db_type': data_types.DOUBLE_PRECISION, 'convert': lambda x: '{}::double precision'.format(x)})
except ImportError:
    pass

try:
    from sqlalchemy.dialects import mysql
    map_data_types.append({'sql_type': mysql.TINYINT, 'map_type': data_types.BOOLEAN, 'db_type': data_types.BOOLEAN})
except ImportError:
    pass

try:
    from sqlalchemy.dialects import mssql
    map_data_types.append({'sql_type': mssql.BIT, 'map_type': data_types.BOOLEAN, 'db_type': data_types.BOOLEAN})
    map_data_types.append({'sql_type': mssql.VARBINARY, 'map_type': data_types.CHAR, 'db_type': data_types.CHAR})
    map_data_types.append({'sql_type': mssql.SMALLDATETIME, 'map_type': data_types.DATE_TIME, 'db_type': data_types.DATE_TIME})
    map_data_types.append({'sql_type': mssql.MONEY, 'map_type': data_types.FLOAT, 'db_type': data_types.MONEY, 'convert': lambda x: '{}::double precision'.format(x)})
    map_data_types.append({'sql_type': mssql.SMALLMONEY, 'map_type': data_types.FLOAT, 'db_type': data_types.MONEY, 'convert': lambda x: '{}::double precision'.format(x)})
except ImportError:
    pass

try:
    from databricks.sqlalchemy.dialect import DatabricksDecimal, DatabricksDate, DatabricksTimestamp
    map_data_types.append({'sql_type': DatabricksDecimal, 'map_type': data_types.FLOAT, 'db_type': data_types.DECIMAL, 'convert': lambda x: '{}::double precision'.format(x)})
    map_data_types.append({'sql_type': DatabricksDate, 'map_type': data_types.DATE, 'db_type': data_types.DATE, 'convert': lambda x: '{}::text::date'.format(x)})
    map_data_types.append({'sql_type': DatabricksTimestamp, 'map_type': data_types.DATE_TIME, 'db_type': data_types.DATE_TIME, 'convert': lambda x: '{}::timestamp with time zone'.format(x)})
except ImportError:
    pass


def sql_to_map_type(value):
    for rule in reversed(map_data_types):
        if is_instance_or_subclass(value, rule['sql_type']):
            return rule['map_type']
    logger.warning('Unknown database type: {}'.format(str(value)))
    return default_map_type


def map_to_sql_type(value):
    for rule in map_data_types:
        if rule['map_type'] == value:
            return rule['sql_type']
    logger.warning('Unknown database type: {}'.format(str(value)))
    return default_sql_type


def sql_to_db_type(value):
    for rule in reversed(map_data_types):
        if is_instance_or_subclass(value, rule['sql_type']):
            return rule['db_type']
    logger.warning('Unknown database type: {}'.format(str(value)))
    return default_db_type


def db_to_sql_type(value):
    for rule in map_data_types:
        if rule['db_type'] == value:
            return rule['sql_type']
    logger.warning('Unknown database type: {}'.format(str(value)))
    return default_sql_type


def get_db_type_convert(value):
    for rule in map_data_types:
        if rule['db_type'] == value:
            return rule.get('convert')
    logger.warning('Unknown database type: {}'.format(str(value)))


def get_sql_type_convert(value):
    for rule in reversed(map_data_types):
        if is_instance_or_subclass(value, rule['sql_type']):
            return rule.get('convert')
    logger.warning('Unknown database type: {}'.format(str(value)))



from jet_bridge_base.responses.base import Response


class TemplateResponse(Response):
    headers = {'Content-Type': 'text/html'}

    def __init__(self, template, data=None, status=None, headers=None, exception=False, content_type=None):
        self.template = template
        super(TemplateResponse, self).__init__(data, status, headers, exception, content_type)



from .common import inspect_uniform, aliased_uniform, get_session_engine
from .db import init_database_connection, load_mapped_base
from .discover import discover_connection, discover_tables
from .metadata_file import dump_metadata_file, remove_metadata_file
from .queryset import desc_uniform, empty_filter, get_queryset_order_by, get_queryset_limit, apply_default_ordering, \
    queryset_count_optimized, queryset_aggregate, queryset_group, get_sql_aggregate_func_by_name, \
    get_sql_group_func_lookup, queryset_search
from .timezones import fetch_default_timezone, apply_session_timezone
from .mongo import MongoDeclarativeMeta, MongoColumn, MongoDesc, MongoQueryset, MongoSession



from .mongo_queryset import MongoQueryset
from .mongo_column import MongoColumn
from .mongo_declarative_meta import MongoDeclarativeMeta


class MongoSession(object):
    info = {}
    records = []

    def __init__(self, db):
        self.db = db

    def commit(self):
        for record in self.records:
            name = record.get_meta().table_name
            update_pending = record.get_update_pending()

            if record.is_delete_pending() and not record.is_create_pending():
                self.db[name].delete_one({'_id': record._id})
            elif record.is_create_pending():
                record_data = record.get_data()
                data = {}

                for key, value in record_data.items():
                    if key == '_id':
                        continue
                    data[key] = value

                self.db[name].insert_one(data)
            elif len(update_pending):
                record_data = record.get_data()
                data = {}

                for key, value in record_data.items():
                    if key not in update_pending:
                        continue
                    data[key] = value

                self.db[name].update_one({'_id': record._id}, {'$set': data})

            record.clear_pending()

    def rollback(self):
        pass

    def close(self):
        pass

    def query(self, *args):
        if isinstance(args[0], MongoDeclarativeMeta):
            name = args[0].get_name()
            select = None
        elif len(args) > 0 and all(map(lambda x: isinstance(x, MongoColumn), args)):
            name = args[0].table.name
            select = args
        else:
            raise Exception('Unsupported query args={}'.format(args))

        return MongoQueryset(self, name, select=select)

    def add(self, record):
        self.bind_record(record)
        record.mark_create()

    def delete(self, record):
        record.mark_delete()

    def bind_record(self, record):
        self.records.append(record)



import threading
import time
from datetime import timedelta

from graphql import GraphQLError

from jet_bridge_base.db import request_connection_cache, get_mapped_base, get_connection_id_short
from jet_bridge_base.exceptions.permission_denied import PermissionDenied
from jet_bridge_base.logger import logger
from jet_bridge_base.permissions import HasProjectPermissions
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.utils.common import get_random_string
from jet_bridge_base.utils.graphql import GraphQLSchemaGenerator
from jet_bridge_base.utils.process import get_memory_usage_human, get_memory_usage
from jet_bridge_base.utils.tables import get_table_name
from jet_bridge_base.utils.track_database import track_database_async
from jet_bridge_base.views.base.api import APIView


class GraphQLView(APIView):
    permission_classes = (HasProjectPermissions,)
    track_queries = True

    def before_dispatch_permissions_check(self, request):
        pass

    def required_project_permission(self, request):
        return {
            'permission_type': 'model',
            'permission_object': request.context.get('model'),
            'permission_actions': 'r'
        }

    def get_track_query_name(self, request):
        return '{} {} {}'.format(request.method, request.path, request.context.get('model'))

    def map_gql_error(self, error):
        if isinstance(error, GraphQLError):
            return error.message
        else:
            return str(error)

    def wait_schema(self, request, schema_key, wait_schema):
        if not wait_schema:
            return

        id_short = get_connection_id_short(request)

        logger.info('[{}] Waiting GraphQL schema "{}"...'.format(id_short, wait_schema['id']))

        generated_condition = wait_schema['generated']
        with generated_condition:
            timeout = timedelta(minutes=10).total_seconds()
            generated_condition.wait(timeout=timeout)

        with request_connection_cache(request) as cache:
            cached_schema = cache.get(schema_key)
            if cached_schema and cached_schema['instance']:
                logger.info('[{}] Found GraphQL schema "{}"'.format(id_short, wait_schema['id']))
                return cached_schema['instance']
            else:
                logger.info('[{}] Not found GraphQL schema "{}"'.format(id_short, wait_schema['id']))

    def create_schema_object(self):
        new_schema_id = get_random_string(32)
        new_schema_generated = threading.Condition()

        return {
            'id': new_schema_id,
            'instance': None,
            'get_schema_time': None,
            'memory_usage_approx': None,
            'generated': new_schema_generated
        }

    def create_schema(self, request, schema_key, new_schema, draft):
        id_short = get_connection_id_short(request)
        memory_usage_before = get_memory_usage()

        try:
            logger.info('[{}] Generating GraphQL schema "{}"...'.format(id_short, new_schema['id']))

            def before_resolve(request, mapper, *args, **kwargs):
                MappedBase = get_mapped_base(request)
                request.context['model'] = get_table_name(MappedBase.metadata, mapper.selectable)
                self.check_permissions(request)

            def on_progress_updated(request, new_schema, current_name, i, total):
                if current_name is not None:
                    logger.info('[{}] Generating GraphQL schema "{}" ({} / {}) (Mem:{})...'.format(
                        id_short,
                        current_name,
                        i + 1,
                        total,
                        get_memory_usage_human()
                    ))

                with request_connection_cache(request) as cache:
                    cached_schema = cache.get(schema_key)

                    if cached_schema and cached_schema['id'] == new_schema['id']:
                        new_schema = {**cached_schema, 'tables_processed': i, 'tables_total': total}
                        cache[schema_key] = new_schema

            get_schema_start = time.time()
            schema = GraphQLSchemaGenerator().get_schema(
                request,
                draft,
                before_resolve=before_resolve,
                on_progress_updated=lambda name, i, total: on_progress_updated(request, new_schema, name, i, total)
            )
            get_schema_end = time.time()
            get_schema_time = round(get_schema_end - get_schema_start, 3)
            memory_usage_approx = get_memory_usage() - memory_usage_before

            with request_connection_cache(request) as cache:
                cached_schema = cache.get(schema_key)

                if cached_schema and cached_schema['id'] == new_schema['id']:
                    new_schema = {
                        **cached_schema,
                        'instance': schema,
                        'get_schema_time': get_schema_time,
                        'memory_usage_approx': memory_usage_approx
                    }
                    cache[schema_key] = new_schema

                    logger.info('[{}] Saved GraphQL schema "{}" (Mem:{})'.format(
                        id_short,
                        new_schema['id'],
                        get_memory_usage_human()
                    ))
                else:
                    logger.info('[{}] Ignoring GraphQL schema result "{}", existing: "{}"'.format(
                        id_short,
                        new_schema['id'],
                        cached_schema.get('id') if cached_schema else None
                    ))

            return schema
        except Exception as e:
            with request_connection_cache(request) as cache:
                cached_schema = cache.get(schema_key)

                if cached_schema and cached_schema['id'] == new_schema['id']:
                    del cache[schema_key]

            raise e
        finally:
            generated_condition = new_schema['generated']
            with generated_condition:
                generated_condition.notify_all()

    def get_schema(self, request, draft):
        schema_key = 'graphql_schema_draft' if draft else 'graphql_schema'
        wait_schema = None

        with request_connection_cache(request) as cache:
            cached_schema = cache.get(schema_key)

            if cached_schema and cached_schema['instance']:
                return cached_schema['instance']
            elif cached_schema and not cached_schema['instance']:
                wait_schema = cached_schema
            else:
                new_schema = self.create_schema_object()
                cache[schema_key] = new_schema

        if wait_schema:
            existing_schema = self.wait_schema(request, schema_key, wait_schema)
            if existing_schema:
                return existing_schema
            else:
                with request_connection_cache(request) as cache:
                    new_schema = self.create_schema_object()
                    cache[schema_key] = new_schema

        return self.create_schema(request, schema_key, new_schema, draft)

    def get(self, request, *args, **kwargs):
        return self.post(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        track_database_async(request)

        draft = bool(request.get_argument('draft', False))
        validate = bool(request.data.get('validate', True))

        if 'query' not in request.data:
            return JSONResponse({})

        try:
            schema = self.get_schema(request, draft)
        except Exception as e:
            return JSONResponse({'errors': ['Failed to get table schema: {}'.format(e)]})

        query = request.data.get('query')
        context_value = {
            'request': request,
            'session': request.session
        }
        result = schema.execute(
            query,
            variables={},
            context_value=context_value,
            validate=validate
        )

        if result.errors is not None and len(result.errors):
            error = result.errors[0]
            if hasattr(error, 'original_error'):
                error = error.original_error
            if isinstance(error, PermissionDenied):
                raise error
            return JSONResponse({'errors': map(lambda x: self.map_gql_error(x), result.errors)})

        response = {
            'data': result.data
        }

        if'graphql_data_query_time' in request.context:
            response['data_query_time'] = request.context['graphql_data_query_time']

        if'graphql_count_query_time' in request.context:
            response['count_query_time'] = request.context['graphql_count_query_time']

        return JSONResponse(response)



from jet_bridge_base.responses.base import Response
from jet_bridge_base.status import HTTP_404_NOT_FOUND


class NotFoundResponse(Response):

    def __init__(self):
        super(NotFoundResponse, self).__init__(
            status=HTTP_404_NOT_FOUND,
            data='<h1>Not Found</h1><p>The requested URL {} was not found on this server.</p>'
        )



from sqlalchemy import Column, String, Boolean, Integer, Sequence, Index

from jet_bridge_base.models.base import Base


class ModelRelationOverrideModel(Base):
    __tablename__ = 'model_relation_override'

    id = Column(Integer, Sequence('id_seq', start=1), primary_key=True)

    connection_id = Column(String)
    model = Column(String)
    draft = Column(Boolean)
    name = Column(String)
    direction = Column(String)
    local_field = Column(String)
    related_model = Column(String)
    related_field = Column(String)

    __table_args__ = (
        Index('model_draft', connection_id, model, draft),
    )

    def __repr__(self):
        return '{} {} {}'.format(self.connection_id, self.model, self.draft)



import datetime
import six
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.sql import sqltypes

from jet_bridge_base import fields
from jet_bridge_base.db import get_default_timezone
from jet_bridge_base.db_types import inspect_uniform, MongoColumn, get_session_engine
from jet_bridge_base.serializers.serializer import Serializer
from jet_bridge_base.utils.exceptions import validation_error_from_database_error
from jet_bridge_base.models import data_types

data_types = [
    {'query': 'VARCHAR', 'operator': 'startswith', 'map_type': data_types.CHAR, 'data_type': fields.CharField},
    {'query': 'TEXT', 'operator': 'equals', 'map_type': data_types.TEXT, 'data_type': fields.CharField},
    {'query': 'BIT', 'operator': 'equals', 'map_type': data_types.BOOLEAN, 'data_type': fields.BooleanField},
    {'query': 'TINYINT', 'operator': 'equals', 'map_type': data_types.SMALL_INTEGER, 'data_type': fields.IntegerField},
    {'query': 'BOOLEAN', 'operator': 'equals', 'map_type': data_types.BOOLEAN, 'data_type': fields.BooleanField},
    {'query': 'INTEGER[]', 'operator': 'startswith', 'map_type': data_types.JSON, 'data_type': fields.ArrayField},
    {'query': 'INTEGER', 'operator': 'equals', 'map_type': data_types.INTEGER, 'data_type': fields.IntegerField},
    {'query': 'SMALLINT', 'operator': 'equals', 'map_type': data_types.SMALL_INTEGER, 'data_type': fields.IntegerField},
    {'query': 'BIGINT', 'operator': 'equals', 'map_type': data_types.BIG_INTEGER, 'data_type': fields.IntegerField},
    {'query': 'FLOAT', 'operator': 'equals', 'map_type': data_types.FLOAT, 'data_type': fields.FloatField},
    {'query': 'DECIMAL', 'operator': 'equals', 'map_type': data_types.DECIMAL, 'data_type': fields.FloatField},
    {'query': 'DOUBLE_PRECISION', 'operator': 'equals', 'map_type': data_types.DOUBLE_PRECISION, 'data_type': fields.FloatField},
    {'query': 'MONEY', 'operator': 'equals', 'map_type': data_types.MONEY, 'data_type': fields.FloatField},
    {'query': 'SMALLMONEY', 'operator': 'equals', 'map_type': data_types.MONEY, 'data_type': fields.FloatField},
    {'query': 'NUMERIC', 'operator': 'startswith', 'map_type': data_types.NUMBER, 'data_type': fields.CharField},
    {'query': 'VARCHAR', 'operator': 'startswith', 'map_type': data_types.CHAR, 'data_type': fields.CharField},
    {'query': 'TIMESTAMP', 'operator': 'startswith', 'map_type': data_types.DATE_TIME, 'data_type': fields.DateTimeField},
    {'query': 'DATETIME', 'operator': 'startswith', 'map_type': data_types.DATE_TIME, 'data_type': fields.DateTimeField},
    {'query': 'JSON', 'operator': 'startswith', 'map_type': data_types.JSON, 'data_type': fields.JSONField},
    {'query': 'ARRAY', 'operator': 'equals', 'map_type': data_types.JSON, 'data_type': fields.JSONField},
    {'query': 'BINARY', 'operator': 'startswith', 'map_type': data_types.BINARY, 'data_type': fields.BinaryField},
    {'query': 'VARBINARY', 'operator': 'startswith', 'map_type': data_types.BINARY, 'data_type': fields.BinaryField},
    {'query': 'geometry', 'operator': 'startswith', 'map_type': data_types.GEOMETRY, 'data_type': fields.WKTField},
    {'query': 'geography', 'operator': 'startswith', 'map_type': data_types.GEOGRAPHY, 'data_type': fields.WKTField},
]
default_data_type = fields.CharField


def get_column_data_type(column):
    if isinstance(column, MongoColumn):
        for rule in reversed(data_types):
            if rule['map_type'] == column.type:
                return rule['data_type']
    else:
        try:
            data_type = six.text_type(column.type)
        except:
            data_type = 'NullType'

        for rule in data_types:
            if rule['operator'] == 'equals' and data_type == rule['query']:
                return rule['data_type']
            elif rule['operator'] == 'startswith' and data_type[:len(rule['query'])] == rule['query']:
                return rule['data_type']

    return default_data_type


class ModelSerializer(Serializer):

    def __init__(self, *args, **kwargs):
        super(ModelSerializer, self).__init__(*args, **kwargs)
        self.session = kwargs.get('context', {}).get('session', None)
        self.model = self.meta.model

    def get_fields(self):
        result = super(ModelSerializer, self).get_fields()

        if hasattr(self.meta, 'model_fields'):
            mapper = inspect_uniform(self.meta.model)
            columns = dict(map(lambda x: (x.key, x), mapper.columns))

            for field_name in self.meta.model_fields:
                column = columns.get(field_name)
                data_type = get_column_data_type(column)
                kwargs = {'context': self.context, 'serializer': self}

                # if column.primary_key and column.autoincrement:
                #     kwargs['read_only'] = True
                if column.autoincrement or column.default or column.server_default or column.nullable:
                    kwargs['required'] = False

                field = data_type(**kwargs)
                field.field_name = field_name
                result.append(field)

        return result

    def prepare_datetime_timezone_naive(self, column, value, default_timezone):
        if not isinstance(value, datetime.datetime) or not value.tzinfo:
            return value

        if not default_timezone:
            return value

        column_type = column.expression.type

        if isinstance(column_type, sqltypes.DATETIME) and not column_type.timezone:
            return value.astimezone(default_timezone)
        else:
            return value

    def prepare_data_timezone_naive(self, data):
        request = self.context.get('request')
        default_timezone = get_default_timezone(request) if request else None

        for key, value in data.items():
            column = getattr(self.model, key, None)

            if not column:
                continue

            data[key] = self.prepare_datetime_timezone_naive(column, value, default_timezone)

    def create_instance(self, validated_data):
        ModelClass = self.meta.model
        return ModelClass(**validated_data)

    def create(self, validated_data):
        mapper = inspect_uniform(self.meta.model)
        primary_key = mapper.primary_key[0]
        primary_key_specified = primary_key.name in validated_data

        if primary_key.autoincrement and primary_key_specified and not validated_data.get(primary_key.name):
            primary_key_specified = False
            del validated_data[primary_key.name]

        self.prepare_data_timezone_naive(validated_data)

        instance = self.create_instance(validated_data)
        self.session.add(instance)

        if primary_key.autoincrement and primary_key_specified:
            if get_session_engine(self.session) == 'postgresql':
                if mapper.selectable.schema:
                    table_name = '"{}"."{}"'.format(mapper.selectable.schema, mapper.selectable.name)
                else:
                    table_name = '"{}"'.format(mapper.selectable.name)

                self.session.execute('''
                    SELECT 
                        pg_catalog.setval(
                            pg_catalog.pg_get_serial_sequence('{}', '{}'), 
                            max("{}")
                        ) 
                    FROM 
                        {}
                '''.format(
                    table_name,
                    primary_key.name,
                    primary_key.name,
                    table_name,
                ))

        try:
            self.session.commit()
        except SQLAlchemyError as e:
            self.session.rollback()
            raise validation_error_from_database_error(e, self.model)

        return instance

    def update(self, instance, validated_data):
        self.prepare_data_timezone_naive(validated_data)

        for attr, value in validated_data.items():
            setattr(instance, attr, value)

        try:
            self.session.commit()
        except SQLAlchemyError as e:
            self.session.rollback()
            raise validation_error_from_database_error(e, self.model)

        return instance



from jet_bridge_base import status
from jet_bridge_base.exceptions.api import APIException


class MissingArgumentError(APIException):
    default_detail = 'Invalid input'
    default_code = 'invalid'
    default_status_code = status.HTTP_400_BAD_REQUEST

    def __init__(self, arg_name):
        super(MissingArgumentError, self).__init__(detail='Missing argument %s' % arg_name)
        self.arg_name = arg_name



from jet_bridge_base.fields.field import Field


class BooleanField(Field):
    TRUE_VALUES = {
        't', 'T',
        'y', 'Y', 'yes', 'YES',
        'true', 'True', 'TRUE',
        'on', 'On', 'ON',
        '1', 1,
        True
    }
    FALSE_VALUES = {
        'f', 'F',
        'n', 'N', 'no', 'NO',
        'false', 'False', 'FALSE',
        'off', 'Off', 'OFF',
        '0', 0, 0.0,
        False
    }

    def to_internal_value_item(self, value):
        if value in self.TRUE_VALUES:
            return True
        elif value in self.FALSE_VALUES:
            return False
        return bool(value)

    def to_representation_item(self, value):
        return value



import psutil

from jet_bridge_base.utils.common import format_size


def get_memory_usage():
    process = psutil.Process()
    return process.memory_info().rss


def get_memory_usage_human():
    memory_used = get_memory_usage()
    return format_size(memory_used)
